import cv2
import numpy as np
from logger import logger
from common_utils.check_utils import check_input_path_and_output_dir, \
    check_file_exists
from common_utils.image_utils import resize_img
from common_utils.path_utils import get_dirpath_from_filepath, get_rootname_from_path, get_extension_from_path
from common_utils.file_utils import delete_file
from ...labelme.annotation import LabelMeAnnotation, LabelMeAnnotationParser, Shape, ShapeHandler
from ...labelme.writer import LabelMeAnnotationWriter
from common_utils.common_types import Point, Rectangle, Polygon, Size, Resize
from common_utils.common_types.bbox import BBox
from ...labelme.writer import LabelMeAnnotationWriter

def write_cropped_image(src_path: str, dst_path: str, bbox: BBox, verbose: bool=False):
    check_input_path_and_output_dir(input_path=src_path, output_path=dst_path)
    img = cv2.imread(src_path)
    int_bbox = bbox.to_int()
    cropped_img = img[int_bbox.ymin:int_bbox.ymax, int_bbox.xmin:int_bbox.xmax, :]
    cv2.imwrite(filename=dst_path, img=cropped_img)

def write_cropped_json(src_img_path: str, src_json_path: str, dst_img_path: str, dst_json_path: str, bound_type='rect', verbose: bool=False):
    def process_shape(shape: Shape, bbox: BBox, new_shape_handler: ShapeHandler):
        points = [Point.from_list(point) for point in shape.points]
        
        contained_count = 0
        for point in points:
            if bbox.contains(point):
                contained_count += 1
        if contained_count == 0:
            return
        elif contained_count == len(points):
            pass
        else:
            logger.error(f"Found a shape that is only partially contained by a bbox.")
            logger.error(f"Shape: {shape}")
            logger.error(f"BBox: {bbox}")

        cropped_points = [Point(x=point.x-bbox.xmin, y=point.y-bbox.ymin) for point in points]
        for point in cropped_points:
            if point.x < 0 or point.y < 0:
                logger.error(f"Encountered negative point after crop: {point}")
                raise Exception
        new_shape = shape.copy()
        new_shape.points = [cropped_point.to_list() for cropped_point in cropped_points]
        new_shape_handler.add(new_shape)
    
    check_input_path_and_output_dir(input_path=src_img_path, output_path=dst_img_path)
    check_input_path_and_output_dir(input_path=src_json_path, output_path=dst_json_path)
    output_img_dir = get_dirpath_from_filepath(dst_img_path)

    annotation = LabelMeAnnotation(
        annotation_path=src_img_path,
        img_dir=dst_img_path,
        bound_type=bound_type
    )
    parser = LabelMeAnnotationParser(annotation_path=src_json_path)
    parser.load()

    bbox_list = []
    for rect in parser.shape_handler.rectangles:
        numpy_array = np.array(rect.points)
        if numpy_array.shape != (2, 2):
            logger.error(f"Encountered rectangle with invalid shape: {numpy_array.shape}")
            logger.error(f"rect: {rect}")
            raise Exception
        xmin, xmax = numpy_array.T[0].min(), numpy_array.T[0].max()
        ymin, ymax = numpy_array.T[1].min(), numpy_array.T[1].max()
        bbox_list.append(BBox.from_list([xmin, ymin, xmax, ymax]))

    img = cv2.imread(src_img_path)
    img_h, img_w = img.shape[:2]

    for i, bbox in enumerate(bbox_list):
        bbox = BBox.buffer(bbox)
        new_shape_handler = ShapeHandler()
        for shape_group in [parser.shape_handler.points, parser.shape_handler.rectangles, parser.shape_handler.polygons]:
            for shape in shape_group:
                process_shape(shape=shape, bbox=bbox, new_shape_handler=new_shape_handler)
        new_shape_list = new_shape_handler.to_shape_list()
        if len(new_shape_list) > 0:
            img_rootname, json_rootname = get_rootname_from_path(dst_img_path), get_rootname_from_path(dst_json_path)
            dst_img_dir, dst_json_dir = get_dirpath_from_filepath(dst_img_path), get_dirpath_from_filepath(dst_json_path)
            dst_img_extension = get_extension_from_path(dst_img_path)
            dst_cropped_img_path = f"{dst_img_dir}/{img_rootname}_{i}.{dst_img_extension}"
            dst_cropped_json_path = f"{dst_json_dir}/{json_rootname}_{i}.json"
            write_cropped_image(src_path=src_img_path, dst_path=dst_cropped_img_path, bbox=bbox, verbose=verbose)
            cropped_labelme_ann = annotation.copy()
            cropped_labelme_ann.annotation_path = dst_cropped_json_path
            cropped_labelme_ann.img_dir = dst_img_dir
            cropped_labelme_ann.img_path = dst_cropped_img_path
            cropped_img = cv2.imread(dst_cropped_img_path)
            cropped_img_h, cropped_img_w = cropped_img.shape[:2]
            cropped_labelme_ann.img_height = cropped_img_h
            cropped_labelme_ann.img_width = cropped_img_w
            cropped_labelme_ann.shapes = new_shape_list
            cropped_labelme_ann.shape_handler = new_shape_handler
            writer = LabelMeAnnotationWriter(cropped_labelme_ann)
            writer.write()
            if verbose:
                logger.info(f"Wrote {dst_cropped_json_path}")

def write_resized_image(input_path: str, output_path: str, target_size: Size, silent: bool=False):
    check_input_path_and_output_dir(input_path=input_path, output_path=output_path)
    img = cv2.imread(input_path)
    resized_img = resize_img(img=img, size=target_size)
    cv2.imwrite(filename=output_path, img=resized_img)
    if not silent:
        logger.info(f"Wrote resized image to {output_path}")

def write_resized_json(
    input_img_path: str, input_json_path: str, output_img_path: str, output_json_path: str,
    target_size: Size, bound_type: str='rect', silent: bool=False
):
    # Note: bound_type doesn't have any significance right now.
    check_input_path_and_output_dir(input_path=input_img_path, output_path=output_img_path)
    check_input_path_and_output_dir(input_path=input_json_path, output_path=output_json_path)

    output_img_dir = get_dirpath_from_filepath(output_img_path)

    annotation = LabelMeAnnotation(
        annotation_path=input_json_path,
        img_dir=output_img_dir,
        bound_type=bound_type
    )
    parser = LabelMeAnnotationParser(annotation_path=input_json_path)
    parser.load()
    img = cv2.imread(filename=input_img_path)
    orig_size = Size.from_cv2_shape(img.shape)
    resize = Resize(old_size=orig_size, new_size=target_size)

    for shape in parser.shape_handler.points:
        shape.points = resize.on_point(
            Point.from_labelme_point_list(shape.points)
        ).to_labelme_format()
    for shape in parser.shape_handler.rectangles:
        shape.points = resize.on_rectangle(
            Rectangle.from_labelme_point_list(shape.points)
        ).to_labelme_format()
    for shape in parser.shape_handler.polygons:
        shape.points = resize.on_polygon(
            Polygon.from_labelme_point_list(shape.points)
        ).to_labelme_format()

    # Update Shapes
    parser.shape_handler2shapes()

    # Get Info From Resized Image
    check_file_exists(output_img_path)
    parser.img_path = output_img_path
    img = cv2.imread(output_img_path)
    parser.img_height, parser.img_width = img.shape[:2]

    annotation.copy_from_parser(
        parser=parser,
        annotation_path=output_json_path,
        img_dir=output_img_dir,
        bound_type='rect'
    )
    
    annotation_writer = LabelMeAnnotationWriter(labelme_annotation=annotation)
    annotation_writer.write()
    if not silent:
        logger.info(f"Wrote resized labelme annotation to {output_json_path}")

def copy_annotation(src_img_path: str, src_json_path: str, dst_img_path: str, dst_json_path: str, bound_type: str):
    src_img_dir = get_dirpath_from_filepath(src_img_path)
    dst_img_dir = get_dirpath_from_filepath(dst_img_path)
    labelme_annotation = LabelMeAnnotation(
        annotation_path=src_json_path,
        img_dir=src_img_dir,
        bound_type=bound_type
    )
    labelme_annotation.load()
    labelme_annotation.img_dir = dst_img_dir
    labelme_annotation.img_path = dst_img_path
    labelme_annotation.annotation_path = dst_json_path
    writer = LabelMeAnnotationWriter(labelme_annotation)
    writer.write()

def move_annotation(src_img_path: str, src_json_path: str, dst_img_path: str, dst_json_path: str, bound_type: str):
    copy_annotation(
        src_img_path=src_img_path,
        src_json_path=src_json_path,
        dst_img_path=dst_img_path,
        dst_json_path=dst_json_path,
        bound_type=bound_type
    )
    delete_file(src_json_path)


from __future__ import annotations

import cv2
from typing import List
import json

from logger import logger
from common_utils.time_utils import get_present_year, get_present_time_Ymd, \
    get_ctime
from common_utils.user_utils import get_username
from common_utils.check_utils import check_required_keys, check_file_exists, check_type_from_list
from common_utils.path_utils import get_filename
from common_utils.file_utils import file_exists
from common_utils.common_types.keypoint import Keypoint2D_List, Keypoint3D_List
from common_utils.common_types.bbox import BBox
from common_utils.common_types.segmentation import Segmentation

from ...camera import Camera
from ....base import BaseStructObject

class COCO_Info(BaseStructObject['COCO_Info']):
    def __init__(
        self,
        description: str,
        url: str="https://github.com/cm107/annotation_utils",
        version: str="1.0",
        year: str=get_present_year(),
        contributor: str=get_username(),
        date_created: str=get_present_time_Ymd()
    ):
        super().__init__()
        self.description = description
        self.url = url
        self.version = version
        self.year = year
        self.contributor = contributor
        self.date_created = date_created

    def __str__(self):
        print_str = ""
        print_str += f"description:\n\t{self.description}\n"
        print_str += f"url:\n\t{self.url}\n"
        print_str += f"version:\n\t{self.version}\n"
        print_str += f"year:\n\t{self.year}\n"
        print_str += f"contributor:\n\t{self.contributor}\n"
        print_str += f"date_created:\n\t{self.date_created}\n"
        return print_str

    @classmethod
    def from_dict(cls, info_dict: dict) -> COCO_Info:
        check_required_keys(
            info_dict,
            required_keys=['description', 'url', 'version', 'year', 'contributor', 'date_created']
        )
        return COCO_Info(
            description=info_dict['description'],
            url=info_dict['url'],
            version=info_dict['version'],
            year=info_dict['year'],
            contributor=info_dict['contributor'],
            date_created=info_dict['date_created']
        )

    @classmethod
    def load_from_path(cls, json_path: str) -> COCO_Info:
        check_file_exists(json_path)
        json_dict = json.load(open(json_path, 'r'))
        return COCO_Info.from_dict(json_dict)

class COCO_License(BaseStructObject['COCO_License']):
    def __init__(self, url: str, id: int, name: str):
        self.url = url
        self.id = id
        self.name = name

    def __str__(self):
        return f"url: {self.url}, id: {self.id}, name: {self.name}"

    def is_equal_to(self, other: COCO_License, exclude_id: bool=True) -> bool:
        result = True
        result = result and self.url == other.url
        result = result and self.name == other.name
        if not exclude_id:
            result = result and self.id == other.id
        return result

    @classmethod
    def from_dict(cls, license_dict: dict) -> COCO_License:
        check_required_keys(
            license_dict,
            required_keys=['url', 'id', 'name']
        )
        return COCO_License(
            url=license_dict['url'],
            id=license_dict['id'],
            name=license_dict['name']
        )

    @classmethod
    def load_from_path(cls, json_path: str) -> COCO_License:
        check_file_exists(json_path)
        json_dict = json.load(open(json_path, 'r'))
        return COCO_License.from_dict(json_dict)

class COCO_Image(BaseStructObject['COCO_License']):
    def __init__(
        self, license_id: int, file_name: str, coco_url: str,
        height: int, width: int, date_captured: str, flickr_url: str, id: int
    ):
        self.license_id = license_id
        self.file_name = file_name
        self.coco_url = coco_url
        self.height = height
        self.width = width
        self.date_captured = date_captured
        self.flickr_url = flickr_url
        self.id = id

    def __str__(self):
        print_str = "========================\n"
        print_str += f"license_id:\n\t{self.license_id}\n"
        print_str += f"file_name:\n\t{self.file_name}\n"
        print_str += f"coco_url:\n\t{self.coco_url}\n"
        print_str += f"height:\n\t{self.height}\n"
        print_str += f"width:\n\t{self.width}\n"
        print_str += f"date_captured:\n\t{self.date_captured}\n"
        print_str += f"flickr_url:\n\t{self.flickr_url}\n"
        print_str += f"id:\n\t{self.id}\n"
        return print_str

    def is_equal_to(
        self, other: COCO_Image,
        exclude_id: bool=True, exclude_date_captured: bool=False
    ) -> bool:
        result = True
        result = result and self.file_name == other.file_name
        result = result and self.coco_url == other.coco_url
        result = result and self.height == other.height
        result = result and self.width == other.width
        result = result and self.flickr_url == other.flickr_url
        if not exclude_date_captured:
            result = result and self.date_captured == other.date_captured
        if not exclude_id:
            result = result and self.id == other.id
            result = result and self.license_id == other.license_id
        return result

    def to_dict(self) -> dict:
        return {
            'license': self.license_id,
            'file_name': self.file_name,
            'coco_url': self.coco_url,
            'height': self.height,
            'width': self.width,
            'date_captured': self.date_captured,
            'flickr_url': self.flickr_url,
            'id': self.id
        }

    @classmethod
    def from_dict(cls, image_dict: dict) -> COCO_Image:
        check_required_keys(
            image_dict,
            required_keys=[
                'license', 'file_name', 'coco_url',
                'height', 'width', 'date_captured',
                'id'
            ]
        )
        return COCO_Image(
            license_id=image_dict['license'],
            file_name=image_dict['file_name'],
            coco_url=image_dict['coco_url'],
            height=image_dict['height'],
            width=image_dict['width'],
            date_captured=image_dict['date_captured'],
            flickr_url=image_dict['flickr_url'] if 'flickr_url' in image_dict else None,
            id=image_dict['id'],
        )

    @classmethod
    def from_img_path(self, img_path: str, license_id: int, image_id: int) -> COCO_Image:
        check_file_exists(img_path)
        img = cv2.imread(img_path)
        img_h, img_w = img.shape[:2]
        return COCO_Image(
            license_id=license_id,
            file_name=get_filename(img_path),
            coco_url=img_path,
            height=img_h,
            width=img_w,
            date_captured=get_ctime(img_path),
            flickr_url=None,
            id=image_id
        )

    @classmethod
    def load_from_path(cls, json_path: str) -> COCO_Image:
        check_file_exists(json_path)
        json_dict = json.load(open(json_path, 'r'))
        return COCO_Image.from_dict(json_dict)

class COCO_Annotation(BaseStructObject['COCO_License']):
    def __init__(
        self,
        id: int, category_id: int, image_id: int, # Standard Required
        segmentation: Segmentation=None, # Standard Optional
        bbox: BBox=None, area: float=None,
        keypoints: Keypoint2D_List=None, num_keypoints: int=None,
        iscrowd: int=0,
        keypoints_3d: Keypoint3D_List=None, camera: Camera=None # Custom Optional
    ):
        # Standard Required
        self.id = id
        self.category_id = category_id
        self.image_id = image_id

        # Standard Optional
        self.segmentation = segmentation if segmentation is not None else Segmentation([])
        if bbox is not None:
            self.bbox = bbox
        else:
            if len(self.segmentation) > 0:
                self.bbox = self.segmentation.to_bbox()
            else:
                logger.error(f'A COCO_Annotation needs to be given either a bbox or a non-empty segmentation at the very least to make a valid annotation.')
                logger.error(f'id: {id}, category_id: {category_id}, image_id: {image_id}')
                raise Exception
        self.area = area
        self.keypoints = keypoints if keypoints is not None else Keypoint2D_List([])
        self.num_keypoints = num_keypoints if num_keypoints is not None else len(self.keypoints)
        self.iscrowd = iscrowd

        # Custom Optional
        self.keypoints_3d = keypoints_3d
        self.camera = camera

    def __str__(self) -> str:
        print_str = 'COCO_Annotation'
        indent = 1
        print_str += '\n'
        print_str += '\t'*indent + f'segmentation: {self.segmentation}'
        print_str += '\n'
        print_str += '\t'*indent + f'num_keypoints: {self.num_keypoints}'
        print_str += '\n'
        print_str += '\t'*indent + f'area: {self.area}'
        print_str += '\n'
        print_str += '\t'*indent + f'iscrowd: {self.iscrowd}'
        print_str += '\n'
        print_str += '\t'*indent + f'keypoints: {self.keypoints}'
        print_str += '\n'
        print_str += '\t'*indent + f'image_id: {self.image_id}'
        print_str += '\n'
        print_str += '\t'*indent + f'bbox: {self.bbox}'
        print_str += '\n'
        print_str += '\t'*indent + f'category_id: {self.category_id}'
        print_str += '\n'
        print_str += '\t'*indent + f'id: {self.id}'
        print_str += '\n'
        print_str += '\t'*indent + f'keypoints_3d: {self.keypoints_3d}'
        print_str += '\n'
        print_str += '\t'*indent + f'camera: {self.camera}'
        return print_str

    def to_dict(self, strict: bool=True) -> dict:
        if strict:
            data_dict = {
                'segmentation': self.segmentation.to_list(demarcation=False),
                'num_keypoints': self.num_keypoints,
                'area': self.area,
                'iscrowd': self.iscrowd,
                'keypoints': self.keypoints.to_list(demarcation=False),
                'image_id': self.image_id,
                'bbox': self.bbox.to_list(output_format='pminsize'),
                'category_id': self.category_id,
                'id': self.id
            }
            if self.keypoints_3d is not None:
                data_dict['keypoints_3d'] = self.keypoints_3d.to_list(demarcation=False)
            if self.camera is not None:
                data_dict['camera_params'] = self.camera.to_dict()
            return data_dict
        else:
            data_dict = {
                'bbox': self.bbox.to_list(output_format='pminsize'),
                'area': self.area,
                'iscrowd': self.iscrowd,
                'image_id': self.image_id,
                'category_id': self.category_id,
                'id': self.id
            }
            if len(self.segmentation) > 0:
                data_dict['segmentation'] = self.segmentation.to_list(demarcation=False)
            if len(self.keypoints) > 0:
                data_dict['keypoints'] = self.keypoints.to_list(demarcation=False)
                data_dict['num_keypoints'] = self.num_keypoints

            if self.keypoints_3d is not None:
                data_dict['keypoints_3d'] = self.keypoints_3d.to_list(demarcation=False)
            if self.camera is not None:
                data_dict['camera_params'] = self.camera.to_dict()
            return data_dict

    def save_to_path(self, save_path: str, overwrite: bool=False, strict: bool=True):
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at save_path: {save_path}')
            raise Exception
        json_dict = self.to_dict(strict=strict)
        json.dump(json_dict, open(save_path, 'w'), indent=2, ensure_ascii=False)

    @classmethod
    def from_dict(cls, ann_dict: dict, strict: bool=True) -> COCO_Annotation:
        if strict:
            check_required_keys(
                ann_dict,
                required_keys=[
                    'segmentation', 'num_keypoints', 'area',
                    'iscrowd', 'keypoints', 'image_id',
                    'bbox', 'category_id', 'id'
                ]
            )
            return COCO_Annotation(
                segmentation=Segmentation.from_list(ann_dict['segmentation'], demarcation=False),
                num_keypoints=ann_dict['num_keypoints'],
                area=ann_dict['area'],
                iscrowd=ann_dict['iscrowd'],
                keypoints=Keypoint2D_List.from_list(ann_dict['keypoints'], demarcation=False),
                image_id=ann_dict['image_id'],
                bbox=BBox.from_list(ann_dict['bbox'], input_format='pminsize'),
                category_id=ann_dict['category_id'],
                id=ann_dict['id'],
                keypoints_3d=Keypoint3D_List.from_list(ann_dict['keypoints_3d'], demarcation=False) if 'keypoints_3d' in ann_dict else None,
                camera=Camera.from_dict(ann_dict['camera_params']) if 'camera_params' in ann_dict else None
            )
        else:
            check_required_keys(
                ann_dict,
                required_keys=[
                    'id', 'category_id', 'image_id'
                ]
            )
            return COCO_Annotation(
                segmentation=Segmentation.from_list(ann_dict['segmentation'], demarcation=False) if 'segmentation' in ann_dict else None,
                num_keypoints=ann_dict['num_keypoints'] if 'num_keypoints' in ann_dict else None,
                area=ann_dict['area'] if 'area' in ann_dict else None,
                iscrowd=ann_dict['iscrowd'] if 'iscrowd' in ann_dict else None,
                keypoints=Keypoint2D_List.from_list(ann_dict['keypoints'], demarcation=False) if 'keypoints' in ann_dict else None,
                image_id=ann_dict['image_id'],
                bbox=BBox.from_list(ann_dict['bbox'], input_format='pminsize') if 'bbox' in ann_dict else None,
                category_id=ann_dict['category_id'],
                id=ann_dict['id'],
                keypoints_3d=Keypoint3D_List.from_list(ann_dict['keypoints_3d'], demarcation=False) if 'keypoints_3d' in ann_dict else None,
                camera=Camera.from_dict(ann_dict['camera_params']) if 'camera_params' in ann_dict else None
            )

    @classmethod
    def load_from_path(cls, json_path: str, strict: bool=True) -> COCO_Annotation:
        check_file_exists(json_path)
        json_dict = json.load(open(json_path, 'r'))
        return COCO_Annotation.from_dict(ann_dict=json_dict, strict=strict)

class COCO_Category(BaseStructObject['COCO_License']):
    def __init__(
        self, id: int, supercategory: str=None, name: str=None, keypoints: List[str]=None, skeleton: List[list]=None
    ):
        # Standard Required
        self.id = id

        # Standard Optional
        if supercategory is None and name is None:
            logger.error(f'Need to provide COCO_Category with either supercategory or name at the very least, but both are None.')
            logger.error(f'id: {id}')
            raise Exception
        supercategory0 = str(supercategory) if type(supercategory) in [int, float] else supercategory
        name0 = str(name) if type(name) in [int, float] else name
        self.supercategory = supercategory0 if supercategory0 is not None else name0
        self.name = name0 if name0 is not None else supercategory0
        self.keypoints = keypoints if keypoints is not None else []
        self.skeleton = skeleton if skeleton is not None else []

    def __str__(self):
        print_str = "========================\n"
        print_str += f"supercategory:\n\t{self.supercategory}\n"
        print_str += f"id:\n\t{self.id}\n"
        print_str += f"name:\n\t{self.name}\n"
        print_str += f"keypoints:\n\t{self.keypoints}\n"
        print_str += f"skeleton:\n\t{self.skeleton}\n"
        return print_str

    def is_equal_to(
        self, other: COCO_Category,
        exclude_id: bool=True
    ) -> bool:
        result = True
        result = result and self.supercategory == other.supercategory
        result = result and self.name == other.name
        result = result and self.keypoints == other.keypoints
        result = result and self.skeleton == other.skeleton
        if not exclude_id:
            result = result and self.id == other.id
        return result

    def to_dict(self, strict: bool=True) -> dict:
        if strict:
            return self.__dict__
        else:
            result_dict = {
                'id': self.id,
                'name': self.name
            }
            if self.supercategory != self.name:
                result_dict['supercategory'] = self.supercategory
            if len(self.keypoints) > 0:
                result_dict['keypoints'] = self.keypoints
            if len(self.skeleton) > 0:
                result_dict['skeleton'] = self.skeleton
            return result_dict

    def save_to_path(self: T, save_path: str, overwrite: bool=False, strict: bool=True):
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at save_path: {save_path}')
            raise Exception
        json_dict = self.to_dict(strict=strict)
        json.dump(json_dict, open(save_path, 'w'), indent=2, ensure_ascii=False)

    @classmethod
    def from_dict(cls, category_dict: dict, strict: bool=True) -> COCO_Category:
        if strict:
            check_required_keys(
                category_dict,
                required_keys=[
                    'supercategory', 'id', 'name',
                    'keypoints', 'skeleton'
                ]
            )
            return COCO_Category(
                supercategory=category_dict['supercategory'],
                id=category_dict['id'],
                name=category_dict['name'],
                keypoints=category_dict['keypoints'],
                skeleton=category_dict['skeleton'],
            )
        else:
            check_required_keys(
                category_dict,
                required_keys=[
                    'id'
                ]
            )
            return COCO_Category(
                id=category_dict['id'],
                supercategory=category_dict['supercategory'] if 'supercategory' in category_dict else None,
                name=category_dict['name'] if 'name' in category_dict else None,
                keypoints=category_dict['keypoints'] if 'keypoints' in category_dict else None,
                skeleton=category_dict['skeleton'] if 'skeleton' in category_dict else None
            )

    @classmethod
    def from_label_skeleton(
        cls, supercategory: str, name: str, id: int,
        label_skeleton: List[list], skeleton_idx_offset: int=0
    ) -> COCO_Category:
        label_list = []
        for [start_label, end_label] in label_skeleton:
            check_type_from_list([start_label, end_label], valid_type_list=[str])
            if start_label not in label_list:
                label_list.append(start_label)
            if end_label not in label_list:
                label_list.append(end_label)
        label_list.sort()
        int_skeleton = []
        for [start_label, end_label] in label_skeleton:
            start_idx = label_list.index(start_label) + skeleton_idx_offset
            end_idx = label_list.index(end_label) + skeleton_idx_offset
            int_skeleton.append([start_idx, end_idx])
        return COCO_Category(
            supercategory=supercategory,
            id=id,
            name=name,
            keypoints=label_list,
            skeleton=int_skeleton
        )

    def get_label_skeleton(self, skeleton_idx_offset: int=0) -> list:
        str_skeleton = []
        for int_bone in self.skeleton:
            bone_start = self.keypoints[int_bone[0]-skeleton_idx_offset] 
            bone_end = self.keypoints[int_bone[1]-skeleton_idx_offset]
            str_bone = [bone_start, bone_end]
            str_skeleton.append(str_bone)
        return str_skeleton

    @classmethod
    def load_from_path(cls, json_path: str, strict: bool=True) -> COCO_Category:
        check_file_exists(json_path)
        json_dict = json.load(open(json_path, 'r'))
        return COCO_Category.from_dict(json_dict, strict=strict)


import base64
import json
from logger import logger
from ..annotation import LabelMeAnnotation, LabelMeAnnotationHandler

class LabelMeAnnotationWriter:
    def __init__(self, labelme_annotation: LabelMeAnnotation):
        self.labelme_annotation = labelme_annotation

    def build_json_dict(self):
        version = self.labelme_annotation.version
        flags = self.labelme_annotation.flags
        shapes = self.labelme_annotation.shapes
        img_path = self.labelme_annotation.img_path
        img_data = open(img_path, 'rb').read()
        img_data = base64.b64encode(img_data).decode('utf-8')
        img_height = self.labelme_annotation.img_height
        img_width = self.labelme_annotation.img_width

        json_dict = {}
        json_dict['version'] = version
        json_dict['flags'] = flags
        json_dict['shapes'] = shapes
        json_dict['imagePath'] = img_path
        json_dict['imageData'] = img_data
        json_dict['imageHeight'] = img_height
        json_dict['imageWidth'] = img_width
        return json_dict

    def write_json_dict(self, json_dict: dict):
        json.dump(json_dict, open(self.labelme_annotation.annotation_path, 'w'))
        logger.info(f"JSON dict has been written to:\n{self.labelme_annotation.annotation_path}")

    def write(self):
        json_dict = self.build_json_dict()
        self.write_json_dict(json_dict)

class LabelMeAnnotationHandlerWriter:
    def __init__(self, labelme_annotation_handler: LabelMeAnnotationHandler):
        self.labelme_annotation_handler = labelme_annotation_handler

    def write_all(self):
        for labelme_annotation in self.labelme_annotation_handler.annotations.values():
            labelme_annotation_writer = LabelMeAnnotationWriter(labelme_annotation)
            labelme_annotation_writer.write()


class BBox:
    def __init__(self, x: float, y: float, width: float, height: float):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

        self.xmin = x
        self.ymin = y
        self.xmax = x + width
        self.ymax = y + height
        self.area = width * height

    def __str__(self):
        return f"BBox: (x,y)=({self.x},{self.y}), (width,height)=({self.width},{self.height})"

    def __repr__(self):
        return self.__str__()

    def get_minmax(self) -> list:
        return [self.xmin, self.ymin, self.xmax, self.ymax]


from __future__ import annotations
from typing import List
import json
import cv2
import numpy as np
from tqdm import tqdm

from logger import logger
from streamer.recorder import Recorder
from common_utils.check_utils import check_required_keys, check_file_exists, \
    check_dir_exists, check_value, check_type_from_list, check_type, \
    check_value_from_list
from common_utils.file_utils import file_exists, make_dir_if_not_exists, \
    get_dir_contents_len, delete_all_files_in_dir, copy_file
from common_utils.adv_file_utils import get_next_dump_path
from common_utils.path_utils import get_filename, get_dirpath_from_filepath, \
    get_extension_from_path, rel_to_abs_path, find_moved_abs_path, \
    get_extension_from_filename
from common_utils.cv_drawing_utils import \
    cv_simple_image_viewer, SimpleVideoViewer, \
    draw_bbox, draw_keypoints, draw_segmentation, draw_skeleton, \
    draw_text_rows_at_point
from common_utils.common_types.point import Point2D_List
from common_utils.common_types.segmentation import Polygon, Segmentation
from common_utils.common_types.bbox import BBox
from common_utils.common_types.keypoint import Keypoint2D, Keypoint2D_List
from common_utils.time_utils import get_ctime
from common_utils.image_utils import scale_to_max, pad_to_max

from .objects import COCO_Info
from .handlers import COCO_License_Handler, COCO_Image_Handler, \
    COCO_Annotation_Handler, COCO_Category_Handler, \
    COCO_License, COCO_Image, COCO_Annotation, COCO_Category
from .misc import KeypointGroup
from ....labelme.refactored import LabelmeAnnotationHandler, LabelmeAnnotation, LabelmeShapeHandler, LabelmeShape
from ....util.utils.coco import COCO_Mapper_Handler
from ....dataset.refactored.config import DatasetConfigCollectionHandler

class COCO_Dataset:
    """
    This is a class that can be thought of as a COCO dataset manipulation tool.

    info: Represents the 'info' section of the COCO annotation file.
            Basic information about the dataset is stored here.
    licenses: Represents the 'licenses' handler of the COCO annotation file.
                If any of the images are licensed, the license information is stored here.
                If none of the images are licensed, you must provide a one license (e.g. MIT License) for indexing purposes.
    images: Represents the 'images' handler of the COCO annotation file.
            All necessary image-related information is stored here.
    annotations: Represents the 'annotations' handler of the COCO annotation file.
                    All annotation data is stored here.
    categories: Represents the 'categories' handler of the COCO annotation file.
                All labeling conventions are specified here.

    To get started with using this class, you could create an empty COCO_Dataset:

        ```python
        # Call Constructor
        dataset = COCO_Dataset.new(description='This is a new test coco dataset')

        # Add a license
        dataset.licenses.append(COCO_License(...))
        # Add a category
        dataset.categories.append(COCO_Category(...))
        # Add Images
        dataset.images.append(COCO_Image(...))
        # Add Annotations
        dataset.annotations.append(COCO_Annotation(...))
        ...
        
        ```
    
    Or you could load an existing COCO dataset from a json file path:
        ```python
        dataset = COCO_Dataset.load_from_path('/path/to/coco/json/file.json')
        ```

    Once you have a complete COCO dataset, you can easily manipulate the dataset as needed:
        ```python
        # Example: Double The Size Of All Bounding Boxes
        for coco_image in dataset.images:
            for coco_ann in dataset.annotations:
                xmin, ymin, xmax, ymax = coco_ann.bbox.to_list()
                bbox_h, bbox_w = coco_ann.bbox.shape()
                xmin = xmin - 0.5 * bbox_w
                ymin = ymin - 0.5 * bbox_h
                xmax = xmax + 0.5 * bbox_w
                ymax = ymax + 0.5 * bbox_h
                xmin = 0 if xmin < 0 else coco_image.width - 1 if xmin >= coco_image.width else xmin
                ymin = 0 if ymin < 0 else coco_image.height - 1 if ymin >= coco_image.height else ymin
                xmax = 0 if xmax < 0 else coco_image.width - 1 if xmax >= coco_image.width else xmax
                ymax = 0 if ymax < 0 else coco_image.height - 1 if ymax >= coco_image.height else ymax
                coco_ann.bbox = BBox(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)
        ```
    
    Finally, you can save your modified COCO dataset to a file as follows:
        ```python
        dataset.save_to_path('/save/path.json')
        ```

    Refer to the methods in this class for information about other functionality.
    """
    def __init__(
        self, info: COCO_Info, licenses: COCO_License_Handler, images: COCO_Image_Handler,
        annotations: COCO_Annotation_Handler, categories: COCO_Category_Handler
    ):
        self.info = info
        self.licenses = licenses
        self.images = images
        self.annotations = annotations
        self.categories = categories

    @classmethod
    def buffer(cls, coco_dataset: COCO_Dataset) -> COCO_Dataset:
        """
        A buffer that will return the same value, but mark the object as a COCO_Dataset object.
        This can be useful if your IDE doesn't recognize the type of your coco dataset object.
        
        coco_dataset: The object that you would like to send through the buffer.
        """
        return coco_dataset

    def copy(self) -> COCO_Dataset:
        """
        Copies the entirety of the COCO Dataset to a new object, which is located at a different
        location in memory.
        """
        return COCO_Dataset(
            info=self.info.copy(),
            licenses=self.licenses.copy(),
            images=self.images.copy(),
            annotations=self.annotations.copy(),
            categories=self.categories.copy()
        )

    @classmethod
    def new(cls, description: str=None) -> COCO_Dataset:
        """
        Create an empty COCO Dataset.
        description (optional): A description of the new dataset that you are creating.
        """
        coco_info = COCO_Info(description=description) if description is not None else COCO_Info()
        return COCO_Dataset(
            info=coco_info,
            licenses=COCO_License_Handler(),
            images=COCO_Image_Handler(),
            annotations=COCO_Annotation_Handler(),
            categories=COCO_Category_Handler()
        )

    def to_dict(self, strict: bool=True) -> dict:
        """
        Converts the COCO_Dataset object to a dictionary format, which is the standard format of COCO datasets.
        """
        return {
            'info': self.info.to_dict(),
            'licenses': self.licenses.to_dict_list(),
            'images': self.images.to_dict_list(),
            'annotations': self.annotations.to_dict_list(strict=strict),
            'categories': self.categories.to_dict_list(strict=strict)
        }

    @classmethod
    def from_dict(cls, dataset_dict: dict, strict: bool=True) -> COCO_Dataset:
        """
        Converts a coco dataset dictionary (the standard COCO format) to a COCO_Dataset class object.
        """
        check_required_keys(
            dataset_dict,
            required_keys=[
                'info', 'licenses', 'images',
                'annotations', 'categories'
            ]
        )
        return COCO_Dataset(
            info=COCO_Info.from_dict(dataset_dict['info']),
            licenses=COCO_License_Handler.from_dict_list(dataset_dict['licenses']),
            images=COCO_Image_Handler.from_dict_list(dataset_dict['images']),
            annotations=COCO_Annotation_Handler.from_dict_list(dataset_dict['annotations'], strict=strict),
            categories=COCO_Category_Handler.from_dict_list(dataset_dict['categories'], strict=strict)
        )

    def auto_fix_img_paths(self, src_container_dir: str, ignore_old_matches: bool=True):
        """
        Not yet implemented
        """
        raise NotImplementedError
        for coco_image in self.images:
            if not file_exists(coco_image.coco_url) or ignore_old_matches:
                fixed_path = find_moved_abs_path(
                    old_path=coco_image.coco_url, container_dir=src_container_dir,
                    get_first_match=False
                )
                if fixed_path is None:
                    logger.error(f"Couldn't any relative path in {coco_image.coco_url} inside of {src_container_dir}")
                    logger.error(f"Suggestion: Try adjusting src_container_dir to contain all required sources.")
                    raise Exception
                if file_exists(fixed_path):
                    coco_image.coco_url = fixed_path

    def move_images(
        self, dst_img_dir: str,
        preserve_filenames: bool=False, overwrite_duplicates: bool=False, update_img_paths: bool=True, overwrite: bool=False,
        show_pbar: bool=True
    ):
        """
        Combines all image directories specified in the coco_url of each coco image in self.images
        to a single image directory.

        dst_img_dir: The directory where you would like to save the combined image set.
        preserve_filenames: If False, unique filenames will be generated so as to not create a filename conflict.
        overwrite_duplicates: Only applicable when preserve_filenames=True.
                              In the event that two images with the same filename are moved to dst_img_dir from
                              two different source folders, an error will be raised if overwrite_duplicates=False.
                              If overwrite_duplicates=True, the second copy will be overwrite the first copy.
        update_img_paths: If True, all coco_url paths specified in self.images will be updated to reflect the new
                          combined image directory.
        overwrite: If True, all files in dst_img_dir will be deleted before copying images into the folder.
        """
        used_img_dir_list = []
        for coco_image in self.images:
            used_img_dir = get_dirpath_from_filepath(coco_image.coco_url)
            if used_img_dir not in used_img_dir_list:
                check_dir_exists(used_img_dir)
                used_img_dir_list.append(used_img_dir)

        if len(used_img_dir_list) == 0:
            logger.error(f"Couldn't parse used_img_dir_list.")
            logger.error(f"Are the coco_url paths in your dataset's image dictionary correct?")
            raise Exception

        make_dir_if_not_exists(dst_img_dir)
        if get_dir_contents_len(dst_img_dir) > 0:
            if overwrite:
                delete_all_files_in_dir(dst_img_dir, ask_permission=False)
            else:
                logger.error(f'dst_img_dir={dst_img_dir} is not empty.')
                logger.error('Please use overwrite=True if you would like to delete the contents before proceeding.')
                raise Exception

        pbar = tqdm(total=len(self.images), unit='image(s)') if show_pbar else None
        pbar.set_description(f'Moving Images...')
        for coco_image in self.images:
            if not preserve_filenames:
                img_extension = get_extension_from_path(coco_image.coco_url)
                dst_img_path = get_next_dump_path(
                    dump_dir=dst_img_dir,
                    file_extension=img_extension
                )
                dst_img_path = rel_to_abs_path(dst_img_path)
            else:
                img_filename = get_filename(coco_image.coco_url)
                dst_img_path = f'{dst_img_dir}/{img_filename}'
                if file_exists(dst_img_path) and not overwrite_duplicates:
                    logger.error(f'Failed to copy {coco_image.coco_url} to {dst_img_dir}')
                    logger.error(f'{img_filename} already exists in destination directory.')
                    logger.error(f'Hint: In order to use preserve_filenames=True, all filenames in the dataset must be unique.')
                    logger.error(
                        f'Suggestion: Either update the filenames to be unique or use preserve_filenames=False' + \
                        f' in order to automatically assign the destination filename.'
                    )
                    raise Exception
            copy_file(src_path=coco_image.coco_url, dest_path=dst_img_path, silent=True)
            if update_img_paths:
                coco_image.coco_url = dst_img_path
                coco_image.file_name = get_filename(dst_img_path)
            if pbar is not None:
                pbar.update(1)
        if pbar is not None:
            pbar.close()

    def save_to_path(self, save_path: str, overwrite: bool=False, strict: bool=True):
        """
        Save this COCO_Dataset object to a json file in the standard COCO format.
        
        save_path: Path of where you would like to save the dataset.
        overwrite: If True, any existing file that exists at save_path will be overwritten.
        """
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at save_path: {save_path}')
            raise Exception
        json_dict = self.to_dict(strict=strict)
        json.dump(json_dict, open(save_path, 'w'), indent=2, ensure_ascii=False)

    @classmethod
    def load_from_path(cls, json_path: str, img_dir: str=None, check_paths: bool=True, strict: bool=True) -> COCO_Dataset:
        """
        Loads a COCO_Dataset object from a COCO json file.

        json_path: Path to the COCO json file that you would like to load.
        img_dir: If not None, you can specify the image directory location of all images in this dataset.
                 Note: This can only be done if all image files are saved to the same directory.
                 Note: In order to create a dataset that has a unified image directory, use self.move_images
        check_paths: If True, all image paths will be checked as the dataset is loaded.
                     An error will be thrown if the corresponding image files do not exist.
        """
        check_file_exists(json_path)
        json_dict = json.load(open(json_path, 'r'))
        dataset = COCO_Dataset.from_dict(json_dict, strict=strict)
        if img_dir is not None:
            check_dir_exists(img_dir)
            for coco_image in dataset.images:
                coco_image.coco_url = f'{img_dir}/{coco_image.file_name}'
        if check_paths:
            for coco_image in dataset.images:
                check_file_exists(coco_image.coco_url)
        return dataset

    def to_labelme(self, priority: str='seg') -> LabelmeAnnotationHandler:
        """
        Convert a COCO_Dataset object to a LabelmeAnnotationHandler.
        The goal of this method is to convert a COCO formatted dataset to a Labelme formatted dataset.

        priority:
            'seg': Use segmentations to bound keypoints
            'bbox': Use bounding boxes to bound keypoints
        """
        check_value(priority, valid_value_list=['seg', 'bbox'])
        handler = LabelmeAnnotationHandler()
        for coco_image in self.images:
            labelme_ann = LabelmeAnnotation(
                img_path=coco_image.coco_url,
                img_h=coco_image.height, img_w=coco_image.width,
                shapes=LabelmeShapeHandler()
            )
            for coco_ann in self.annotations.get_annotations_from_imgIds([coco_image.id]):
                coco_cat = self.categories.get_obj_from_id(coco_ann.category_id)
                bbox_contains_seg = coco_ann.segmentation.within(coco_ann.bbox)
                if bbox_contains_seg and priority == 'seg':
                    for polygon in coco_ann.segmentation:
                        if len(polygon.to_list(demarcation=True)) < 3:
                            continue
                        labelme_ann.shapes.append(
                            LabelmeShape(
                                label=coco_cat.name,
                                points=Point2D_List.from_list(polygon.to_list(demarcation=True)),
                                shape_type='polygon'
                            )
                        )
                else:
                    labelme_ann.shapes.append(
                        LabelmeShape(
                            label=coco_cat.name,
                            points=coco_ann.bbox.to_point2d_list(),
                            shape_type='rectangle'
                        )
                    )
                if len(coco_ann.keypoints) > 0:
                    for i, kpt in enumerate(coco_ann.keypoints):
                        if kpt.visibility == 0:
                            continue
                        labelme_ann.shapes.append(
                            LabelmeShape(
                                label=coco_cat.keypoints[i],
                                points=Point2D_List.from_list([kpt.point.to_list()]),
                                shape_type='point'
                            )
                        )
            handler.append(labelme_ann)
        return handler

    @classmethod
    def from_labelme(
        cls, labelme_handler: LabelmeAnnotationHandler,
        categories: COCO_Category_Handler,
        img_dir: str=None, remove_redundant: bool=True,
        ensure_no_unbounded_kpts: bool=True,
        ensure_valid_shape_type: bool=True,
        ignore_unspecified_categories: bool=False,
        license_url: str='https://github.com/cm107/annotation_utils/blob/master/LICENSE',
        license_name: str='MIT License'
    ) -> COCO_Dataset:
        """
        Used to convert a LabelmeAnnotationHandler object to a COCO_Dataset object.
        This is meant to be used for converting a labelme dataset to a COCO dataset.

        labelme_handler: LabelmeAnnotationHandler object
        categories: COCO_Category_Handler object
        img_dir: Directory where all of the labelme dataset images are saved.
        remove_redundant: Remove bounding boxes that are contained withing segmentations and vice versa.
        ensure_no_unbounded_kpts: Ensures that all keypoints are bounded by either a bounding_box or segmentation.
        ensure_valid_shape_type: Ensures that all shape_type's in the LabelmeAnnotationHandler are valid.
        ignore_unspecified_categories: If true, all labels that are not specified in categories is ignored.
        license_url: The url of the license that you would like to associate with this converted dataset.
        license_name: The name of the license that is associated with this dataset.
        """
        dataset = COCO_Dataset.new(description='COCO Dataset converted from Labelme using annotation_utils')
        
        # Add a license to COCO Dataset
        dataset.licenses.append( # Assume that the dataset is free to use.
            COCO_License(
                url=license_url,
                name=license_name,
                id=0
            )
        )

        # Make sure at least one category is provided
        if type(categories) is list:
            categories = COCO_Category_Handler(category_list=categories)
        check_type(categories, valid_type_list=[COCO_Category_Handler])
        if len(categories) == 0:
            logger.error(f'Need to provide at least one COCO_Category for conversion to COCO format.')
            raise Exception
        category_names = [category.name for category in categories]

        # Add categories to COCO Dataset
        dataset.categories = categories

        for labelme_ann in labelme_handler:
            img_filename = get_filename(labelme_ann.img_path)
            if img_dir is not None:
                img_path = f'{img_dir}/{img_filename}'
            else:
                img_path = labelme_ann.img_path
            check_file_exists(img_path)
            
            kpt_label2points_list = {}
            bound_group_list = []
            poly_list = []
            poly_label_list = []
            bbox_list = []
            bbox_label_list = []

            if ensure_valid_shape_type:
                for shape in labelme_ann.shapes:
                    check_value(shape.shape_type, valid_value_list=['point', 'polygon', 'rectangle'])
            
            # Gather all segmentations
            for shape in labelme_ann.shapes:
                if shape.shape_type == 'polygon':
                    if shape.label not in category_names:
                        if ignore_unspecified_categories:
                            continue
                        else:
                            logger.error(f'shape.label={shape.label} does not exist in provided categories.')
                            logger.error(f'category_names: {category_names}')
                            raise Exception
                    poly_list.append(
                        Polygon.from_point2d_list(shape.points)
                    )
                    poly_label_list.append(shape.label)
            # Gather all bounding boxes
            for shape in labelme_ann.shapes:
                if shape.shape_type == 'rectangle':
                    if shape.label not in category_names:
                        if ignore_unspecified_categories:
                            continue
                        else:
                            logger.error(f'shape.label={shape.label} does not exist in provided categories.')
                            logger.error(f'category_names: {category_names}')
                            raise Exception
                    bbox_list.append(
                        BBox.from_point2d_list(shape.points)
                    )
                    bbox_label_list.append(shape.label)
            if remove_redundant:
                # Remove segmentation/bbox redundancies
                for poly in poly_list:
                    for i, [bbox, bbox_label] in enumerate(zip(bbox_list, bbox_label_list)):
                        if poly.contains(bbox):
                            del bbox_list[i]
                            del bbox_label_list[i]
                for bbox in bbox_list:
                    for i, [poly, poly_label] in enumerate(zip(poly_list, poly_label_list)):
                        if bbox.contains(poly):
                            del poly_list[i]
                            del poly_label_list[i]
            # Gather all keypoints
            for shape in labelme_ann.shapes:
                if shape.shape_type == 'point':
                    if shape.label not in kpt_label2points_list:
                        kpt_label2points_list[shape.label] = [shape.points[0]]
                    else:
                        kpt_label2points_list[shape.label].append(shape.points[0])

            # Group keypoints inside of polygon bounds
            for poly, poly_label in zip(poly_list, poly_label_list):
                coco_cat = dataset.categories.get_unique_category_from_name(poly_label)
                bound_group = KeypointGroup(bound_obj=poly, coco_cat=coco_cat)
                # Register the keypoints inside of each polygon
                for label, kpt_list in kpt_label2points_list.items():
                    for i, kpt in enumerate(kpt_list):
                        if kpt.within(poly):
                            bound_group.register(kpt=Keypoint2D(point=kpt, visibility=2), label=label)
                            del kpt_label2points_list[label][i]
                            if len(kpt_label2points_list[label]) == 0:
                                del kpt_label2points_list[label]
                            break
                bound_group_list.append(bound_group)
            # Group keypoints inside of bbox bounds
            for bbox, bbox_label in zip(bbox_list, bbox_label_list):
                coco_cat = dataset.categories.get_unique_category_from_name(bbox_label)
                bound_group = KeypointGroup(bound_obj=bbox, coco_cat=coco_cat)
                # Register the keypoints inside of each bounding box
                temp_dict = kpt_label2points_list.copy()
                for label, kpt_list in temp_dict.items():
                    for i, kpt in enumerate(kpt_list):
                        if kpt.within(bbox):
                            bound_group.register(kpt=Keypoint2D(point=kpt, visibility=2), label=label)
                            del kpt_label2points_list[label][i]
                            if len(kpt_label2points_list[label]) == 0:
                                del kpt_label2points_list[label]
                            break
                bound_group_list.append(bound_group)

            if ensure_no_unbounded_kpts:
                # Ensure that there are no leftover keypoints that are unbounded.
                # (This case often results from mistakes during annotation creation.)
                if len(kpt_label2points_list) > 0:
                    logger.error(f'The following keypoints were left unbounded:\n{kpt_label2points_list}')
                    logger.error(f'Image filename: {img_filename}')
                    raise Exception

            if len(bound_group_list) > 0:
                image_id = len(dataset.images)
                # Add image to COCO dataset images
                dataset.images.append(
                    COCO_Image(
                        license_id=0,
                        file_name=get_filename(img_path),
                        coco_url=img_path,
                        height=labelme_ann.img_h,
                        width=labelme_ann.img_w,
                        date_captured=get_ctime(img_path),
                        flickr_url=None,
                        id=image_id
                    )
                )

                # Add segmentation and/or bbox to COCO dataset annotations together with bounded keypoints
                for bound_group in bound_group_list:
                    keypoints = Keypoint2D_List()
                    for label in bound_group.coco_cat.keypoints:
                        label_found = False
                        for kpt, kpt_label in zip(bound_group.kpt_list, bound_group.kpt_label_list):
                            if kpt_label == label:
                                label_found = True
                                keypoints.append(kpt)
                                break
                        if not label_found:
                            keypoints.append(Keypoint2D.from_list([0, 0, 0]))
                    if type(bound_group.bound_obj) is Polygon:
                        bbox = bound_group.bound_obj.to_bbox()
                        dataset.annotations.append(
                            COCO_Annotation(
                                segmentation=Segmentation(polygon_list=[bound_group.bound_obj]),
                                num_keypoints=len(bound_group.coco_cat.keypoints),
                                area=bbox.area(),
                                iscrowd=0,
                                keypoints=keypoints,
                                image_id=image_id,
                                bbox=bbox,
                                category_id=bound_group.coco_cat.id,
                                id=len(dataset.annotations)
                            )
                        )
                    elif type(bound_group.bound_obj) is BBox:
                        dataset.annotations.append(
                            COCO_Annotation(
                                segmentation=Segmentation(polygon_list=[]),
                                num_keypoints=len(bound_group.coco_cat.keypoints),
                                area=bound_group.bound_obj.area(),
                                iscrowd=0,
                                keypoints=keypoints,
                                image_id=image_id,
                                bbox=bound_group.bound_obj,
                                category_id=bound_group.coco_cat.id,
                                id=len(dataset.annotations)
                            )
                        )
                    else:
                        raise Exception

        return dataset

    def update_img_dir(self, new_img_dir: str, check_paths: bool=True):
        """
        This method is used to update the image directory of all images in the dataset.
        Note that this assumes that all images are contained in the same directory.

        new_img_dir: The new image directory that all of your images are contained in.
        check_paths: When True, all image file paths are checked.
                     An error is thrown if an image cannot be found.
        """
        if check_paths:
            check_dir_exists(new_img_dir)

        for coco_image in self.images:
            coco_image.coco_url = f'{new_img_dir}/{coco_image.file_name}'
            if check_paths:
                check_file_exists(coco_image.coco_url)

    @classmethod
    def combine(cls, dataset_list: List[COCO_Dataset], img_dir_list: List[str]=None, show_pbar: bool=False) -> COCO_Dataset:
        """
        Combines a list of COCO_Dataset's into a single COCO_Dataset.

        dataset_list: A list of all of the COCO_Dataset objects that you would like to combine.
        img_dir_list: A list of all of the image directory paths that correspond to each COCO_Dataset in dataset_list.
        show_pbar: If True, a progress bar will be shown while the datasets are combined.
        """
        if img_dir_list is not None:
            if len(img_dir_list) != len(dataset_list):
                logger.error(f'len(img_dir_list) == {len(img_dir_list)} != {len(dataset_list)} == len(dataset_list)')
                raise Exception
            update_img_pbar = tqdm(total=len(img_dir_list), unit='dataset(s)') if show_pbar else None
            if update_img_pbar is not None:
                pbar.set_description(f'Updating Image Paths...')
            for img_dir, dataset in zip(img_dir_list, dataset_list):
                dataset = COCO_Dataset.buffer(dataset)
                dataset.update_img_dir(new_img_dir=img_dir, check_paths=True)
                if update_img_pbar is not None:
                    update_img_pbar.update(1)
            if update_img_pbar is not None:
                update_img_pbar.close()
        
        result_dataset = COCO_Dataset.new(
            description='A combination of many COCO datasets using annotation_utils'
        )
        map_handler = COCO_Mapper_Handler()
        merge_pbar = tqdm(total=len(dataset_list), unit='dataset(s)') if show_pbar else None
        if merge_pbar is not None:
            merge_pbar.set_description(f'Merging Datasets...')
        for i, dataset in enumerate(dataset_list):
            # Process Licenses
            for coco_license in dataset.licenses:
                already_exists = False
                for existing_license in result_dataset.licenses:
                    if coco_license.is_equal_to(existing_license, exclude_id=True):
                        already_exists = True
                        map_handler.license_mapper.add(
                            unique_key=i, old_id=coco_license.id, new_id=existing_license.id
                        )
                        break
                if not already_exists:
                    new_license = coco_license.copy()
                    new_license.id = len(result_dataset.licenses)
                    map_handler.license_mapper.add(
                        unique_key=i, old_id=coco_license.id, new_id=new_license.id
                    )
                    result_dataset.licenses.append(new_license)

            # Process Images
            for coco_image in dataset.images:
                check_file_exists(coco_image.coco_url)
                already_exists = False
                for existing_image in result_dataset.images:
                    if coco_image.is_equal_to(existing_image, exclude_id=True):
                        already_exists = True
                        map_handler.image_mapper.add(
                            unique_key=i, old_id=coco_image.id, new_id=existing_image.id
                        )
                        break
                if not already_exists:
                    new_image = coco_image.copy()
                    new_image.id = len(result_dataset.images)
                    map_handler.image_mapper.add(
                        unique_key=i, old_id=coco_image.id, new_id=new_image.id
                    )
                    found, new_image.license_id = map_handler.license_mapper.get_new_id(
                        unique_key=i, old_id=coco_image.license_id
                    )
                    if not found:
                        logger.error(f"Couldn't find license map using unique_key={i}, old_id={coco_image.license_id}")
                        raise Exception
                    result_dataset.images.append(new_image)

            # Process Categories
            for coco_category in dataset.categories:
                already_exists = False
                for existing_category in result_dataset.categories:
                    if coco_category.is_equal_to(existing_category, exclude_id=True):
                        already_exists = True
                        map_handler.category_mapper.add(
                            unique_key=i, old_id=coco_category.id, new_id=existing_category.id
                        )
                        break
                if not already_exists:
                    new_category = coco_category.copy()
                    new_category.id = len(result_dataset.categories)
                    map_handler.category_mapper.add(
                        unique_key=i, old_id=coco_category.id, new_id=new_category.id
                    )
                    result_dataset.categories.append(new_category)

            # Process Annotations
            for coco_ann in dataset.annotations:
                new_ann = coco_ann.copy()
                new_ann.id = len(result_dataset.annotations)
                found, new_ann.image_id = map_handler.image_mapper.get_new_id(
                    unique_key=i, old_id=coco_ann.image_id
                )
                if not found:
                    logger.error(f"Couldn't find image map using unique_key={i}, old_id={coco_ann.image_id}")
                    raise Exception
                found, new_ann.category_id = map_handler.category_mapper.get_new_id(
                    unique_key=i, old_id=coco_ann.category_id
                )
                if not found:
                    logger.error(f"Couldn't find category map using unique_key={i}, old_id={coco_ann.category_id}")
                    raise Exception
                result_dataset.annotations.append(new_ann)
            if merge_pbar is not None:
                merge_pbar.update(1)
        if merge_pbar is not None:
            merge_pbar.close()

        return result_dataset

    @classmethod
    def combine_from_config(cls, config_path: str, img_sort_attr_name: str=None, show_pbar: bool=False) -> COCO_Dataset:
        """
        This is the same as COCO_Dataset.combine, but with this method you don't have to construct each dataset manually.
        Instead, you can just provide a dataset configuration file that specifies the location of all of your coco json files
        ase well as their corresponding image directories.
        For more information about how to make this dataset configuration file, please refer to the DatasetConfigCollectionHandler class.

        config_path: The path to your dataset configuration file.
        img_sort_attr_name: The attribute name that you would like to sort the dataset images by before the datasets are combined.
                            (Example: img_sort_attr_name='file_name')
        show_pbar: If True, a progress bar will be shown while the images and annotations are loaded into the dataset.
        """

        dataset_path_config = DatasetConfigCollectionHandler.load_from_path(config_path)
        config_list = []
        for collection in dataset_path_config:
            for config in collection:
                check_value(config.ann_format, valid_value_list=['coco'])
                config_list.append(config)
        # dataset_dir_list, img_dir_list, ann_path_list, ann_format_list = dataset_path_config.get_paths()
        # check_value_from_list(item_list=ann_format_list, valid_value_list=['coco'])
        dataset_list = []
        # pbar = tqdm(total=len(img_dir_list), unit='dataset(s)') if show_pbar else None
        pbar = tqdm(total=len(config_list), unit='dataset(s)') if show_pbar else None
        if pbar is not None:
            pbar.set_description(f'Loading Dataset List...')
        # for img_dir, ann_path in zip(img_dir_list, ann_path_list):
        for config in config_list:
            # dataset = COCO_Dataset.load_from_path(json_path=ann_path, img_dir=img_dir, check_paths=True)
            dataset = COCO_Dataset.load_from_path(json_path=config.ann_path, img_dir=config.img_dir, check_paths=True)
            if img_sort_attr_name is not None:
                dataset.images.sort(attr_name=img_sort_attr_name)
            dataset_list.append(dataset)
            if pbar is not None:
                pbar.update(1)
        if pbar is not None:
            pbar.close()
        return COCO_Dataset.combine(dataset_list, show_pbar=show_pbar)

    def split(
        self, dest_dir: str,
        split_dirname_list: List[str]=['train', 'test', 'val'], ratio: list=[2, 1, 0], coco_filename_list: List[str]=None,
        shuffle: bool=True, preserve_filenames: bool=False, overwrite: bool=False
    ) -> List[COCO_Dataset]:
        """
        Use this method to split a single coco dataset into multiple datasets.
        This can be useful for when you need to split a single coco dataset into train, test, and validation datasets.

        dest_dir: Path to the folder where you would like to save the split datasets.
        split_dirname_list: Specify the folder names of each split dataset part that you would like to generate.
                            These folders are created in the folder specified by dest_dir.
                            For example, if you want to split your dataset into a train dataset and validation dataset,
                            use split_dirname_list=['train', 'val']
        ratio: Specify the ratio of images between each split dataset part.
               For example, if you would like to have twice as many train images as validation images, use ratio=[2, 1].
        coco_filename_list: Specify the filenames of the coco annotation file of each split dataset part.
                            Example: ['train.json', 'val.json']
        shuffle: If True, the images handler will be shuffled before splitting the dataset into parts.
        preserve_filenames: If True, the image filenames will be preserved during the split.
                            However, if a filename conflict is encountered as a result, an error will be thrown.
                            If False, the image filenames of each dataset part will be automatically generated
                            so as to avoid filename conflicts.
        overwrite: If True, the contents of dest_dir will be deleted before creating a new split dataset folder.
                   If False, an error will be thrown if dest_dir contains any files or directories.
        """

        # Checks
        check_type_from_list([split_dirname_list, ratio, coco_filename_list], valid_type_list=[list])
        if len(split_dirname_list) != len(ratio):
            logger.error(f'len(split_dirname_list) == {len(split_dirname_list)} != {len(ratio)} == len(ratio)')
            raise Exception
        check_type_from_list(split_dirname_list, valid_type_list=[str])
        check_type_from_list(ratio, valid_type_list=[int])
        if coco_filename_list is None:
            coco_filename_list = ['output.json'] * len(split_dirname_list)
        else:
            check_type_from_list(coco_filename_list, valid_type_list=[str])
            if len(coco_filename_list) != len(split_dirname_list):
                logger.error(f'len(coco_filename_list) == {len(coco_filename_list)} != {len(split_dirname_list)} == len(split_dirname_list)')
                raise Exception

        # Prepare Output Directory
        split_dirpath_list = [f'{dest_dir}/{split_dirname}' for split_dirname in split_dirname_list]
        split_imgdir_list = [f'{split_dirpath}/img' for split_dirpath in split_dirpath_list]
        split_cocodir_list = [f'{split_dirpath}/coco' for split_dirpath in split_dirpath_list]
        split_cocopath_list = [f'{split_cocodir}/{coco_filename}' for split_cocodir, coco_filename in zip(split_cocodir_list, coco_filename_list)]
        make_dir_if_not_exists(dest_dir)
        for split_dirpath, split_imgdir, split_cocodir in zip(split_dirpath_list, split_imgdir_list, split_cocodir_list):
            make_dir_if_not_exists(split_dirpath)
            if get_dir_contents_len(split_dirpath) > 0:
                if overwrite:
                    delete_all_files_in_dir(split_dirpath, ask_permission=False)
                else:
                    logger.error(f'Files/Directories were found in: {split_dirpath}')
                    logger.error('Use overwrite=True to overwrite all contents.')
                    raise Exception
            make_dir_if_not_exists(split_imgdir)
            make_dir_if_not_exists(split_cocodir)
            

        # Split COCO Images Into Samples
        locations = np.cumsum([val*int(len(self.images)/sum(ratio)) for val in ratio]) - 1
        start_location = None
        end_location = 0
        count = 0
        coco_image_samples = []
        if shuffle:
            self.images.shuffle()
        while count < len(locations):
            start_location = end_location
            end_location = locations[count]
            count += 1
            coco_image_samples.append(self.images[start_location:end_location].copy())

        # Construct New Datasets
        dataset_list = []
        for coco_image_list, split_dirname, split_imgdir, split_cocopath in \
            tqdm(zip(coco_image_samples, split_dirname_list, split_imgdir_list, split_cocopath_list), total=len(split_dirname_list), unit='part(s)', leave=True):
            dataset = COCO_Dataset.new(description=f'Split {split_dirname} Dataset')
            used_license_id_list = []
            used_category_id_list = []
            for coco_image0 in tqdm(coco_image_list, total=len(coco_image_list), unit='image(s)', leave=False):
                coco_image = coco_image0.copy()
                # Map Image Index
                coco_image = COCO_Image.buffer(coco_image)
                anns = self.annotations.get_annotations_from_imgIds([coco_image.id]).copy()
                new_image_id = len(dataset.images)
                
                # Copy Image
                old_img_path = coco_image.coco_url
                if not preserve_filenames:
                    img_extension = get_extension_from_filename(coco_image.file_name)
                    new_img_path = get_next_dump_path(dump_dir=split_imgdir, file_extension=img_extension)
                else:
                    new_img_path = f'{split_imgdir}/{coco_image.file_name}'
                if file_exists(new_img_path):
                    logger.error(f'Copy failed. Image already exists in destination directory: {new_img_path}')
                    logger.error(f'This is likely because the filenames in your dataset are not unique.')
                    logger.error(f'Use preserve_filenames=False to use automatically generated filenames.')
                    raise Exception
                copy_file(src_path=old_img_path, dest_path=new_img_path, silent=True)

                # Update COCO Image
                coco_image.id = new_image_id
                coco_image.coco_url = new_img_path
                coco_image.file_name = get_filename(new_img_path)
                dataset.images.append(coco_image)
                if coco_image.license_id not in used_license_id_list:
                    used_license_id_list.append(coco_image.license_id)

                for coco_ann0 in anns:
                    coco_ann = coco_ann0.copy()
                    # Map Annotation Index
                    new_ann_id = len(dataset.annotations)

                    # Update COCO Annotation
                    coco_ann.id = new_ann_id
                    coco_ann.image_id = new_image_id
                    dataset.annotations.append(coco_ann)
                    if coco_ann.category_id not in used_category_id_list:
                        used_category_id_list.append(coco_ann.category_id)

            # Add Used Licenses To Dataset and Update Ids
            for coco_license0 in self.licenses:
                coco_license = coco_license0.copy()
                if coco_license.id in used_license_id_list:
                    new_license_id = len(dataset.licenses)
                    for coco_image in dataset.images:
                        if coco_image.license_id == coco_license.id:
                            coco_image.license_id = new_license_id
                    coco_license.id = new_license_id
                    dataset.licenses.append(coco_license)
            
            # Add Used Categories To Dataset and Update Ids
            for coco_cat0 in self.categories:
                coco_cat = coco_cat0.copy()
                if coco_cat.id in used_category_id_list:
                    new_cat_id = len(dataset.categories)
                    for coco_ann in dataset.annotations:
                        if coco_ann.category_id == coco_cat.id:
                            coco_ann.category_id = new_cat_id
                    coco_cat.id = new_cat_id
                    dataset.categories.append(coco_cat)
            
            # Append Dataset To Split Dataset List
            dataset.save_to_path(save_path=split_cocopath, overwrite=False)
            dataset_list.append(dataset)
        return dataset_list

    def prune_keypoints(self, min_num_kpts: int, verbose: bool=False):
        """Used to prune out all of the annotations and images that contain below a certain level of keypoints, which is specified by min_num_kpts.
        
        Arguments:
            min_num_kpts {int} -- [Specify the mininum number of keypoints that can be allowed without the annotation being pruned.]
        
        Keyword Arguments:
            verbose {bool} -- [If True, prints some detailed information to the terminal.] (default: {False})

        Note:
            Using prune_keypoints by itself will not make any changes to image files or annotation files.
            It is merely pruning keypoints within the COCO_Dataset instance.
            In order to delete unnecessary image files and update the annotation file, using the following lines of code:
                dataset.prune_keypoints(min_num_kpts=[INT], verbose=True)
                dataset.move_images(dst_img_dir='/path/to/new/img_dir', preserve_filenames=True, update_img_paths=True, overwrite=True, show_pbar=True)
                dataset.save_to_path(save_path='/path/to/new/ann.json', overwrite=True)
            With this the new dataset is saved to a different location.
            In order to avoid accidently deleting files from the python script, please delete the old dataset files manually.
        """
        ann_idx_list = list(range(len(self.annotations)))
        ann_idx_list.reverse()
        for i in ann_idx_list:
            coco_ann = self.annotations[i]
            visibility_list = [kpt.visibility for kpt in coco_ann.keypoints]
            num_visible = 0
            for v in visibility_list:
                if v > 0.0:
                    num_visible += 1
            if num_visible < min_num_kpts:
                del self.annotations[i]
                if verbose:
                    logger.info(f'Deleted ann id: {coco_ann.id}')
        img_idx_list = list(range(len(self.images)))
        img_idx_list.reverse()
        for i in img_idx_list:
            coco_image = self.images[i]
            coco_anns = self.annotations.get_annotations_from_imgIds([coco_image.id])
            if len(coco_anns) == 0:
                coco_image = self.images[i]
                del self.images[i]
                if verbose:
                    logger.info(f'Deleted image id: {coco_image.id}')

    def remove_categories_by_name(self, category_names: List[str], verbose: bool=False):
        self.categories.remove_by_name(
            names=category_names,
            ann_handler=self.annotations,
            img_handler=self.images,
            license_handler=self.licenses,
            verbose=verbose
        )

    def print_handler_lengths(self):
        logger.info(f'len(licenses): {len(self.licenses)}')
        logger.info(f'len(images): {len(self.images)}')
        logger.info(f'len(annotations): {len(self.annotations)}')
        logger.info(f'len(categories): {len(self.categories)}')

    def draw_annotation(
        self, img: np.ndarray, ann_id: int,
        draw_order: list=['seg', 'bbox', 'skeleton', 'kpt'],
        bbox_color: list=[0, 255, 255], bbox_thickness: list=2, # BBox
        bbox_show_label: bool=True, bbox_label_thickness: int=None,
        bbox_label_only: bool=False,
        seg_color: list=[255, 255, 0], seg_transparent: bool=True, # Segmentation
        kpt_radius: int=4, kpt_color: list=[0, 0, 255], # Keypoints
        show_kpt_labels: bool=True, kpt_label_thickness: int=1,
        kpt_label_only: bool=False, ignore_kpt_idx: list=[],
        kpt_idx_offset: int=0,
        skeleton_thickness: int=5, skeleton_color: list=[255, 0, 0], # Skeleton
        details_corner_pos_ratio: float=0.02, details_height_ratio: float=0.10, # Details
        details_leeway: float=0.4, details_color: list=[255, 0, 255],
        details_thickness: int=2,
        show_bbox: bool=True, show_kpt: bool=True, # Show Flags
        show_skeleton: bool=True, show_seg: bool=True,
        show_details: bool=False
    ) -> np.ndarray:
        """
        Draws the annotation corresponding to ann_id on a given image.

        img: The image array that you would like to draw the annotation on.
        ann_id: The id that corresponds to the annotation that you would like to draw.
        draw_order: The order in which you would like to draw (render) the annotations.
                    Example: If you specify 'bbox' after 'seg', the bounding box will be
                    drawn after the segmentation is drawn.
        bbox_color: The color of the bbox that is to be drawn.
        bbox_thickness: The thickness of the bbox that is to be drawn.
        bbox_show_label: If True, the label of the bbox will be drawn directly above it.
        bbox_label_thickness: The thickness of the bbox label in the event that it is drawn.
        bbox_label_only: If you would rather not draw the bounding box and only show the label,
                         set this to True.
        seg_color: The color of the segmentation that is to be drawn.
        seg_transparent: If True, the segmentation that is drawn will be transparent.
                         If False, the segmentation will be a solid color.
        kpt_radius: The radius of the keypoints that are to be drawn.
        kpt_color: The color of the keypoints that are to be drawn.
        show_kpt_labels: If True, the labels of the keypoints will be drawn directly above each
                         keypoint.
        kpt_label_thickness: The thickness of the keypoint labels in the event that they are drawn.
        kpt_label_only: If True, the keypoints will not be drawn and only the keypoint labels will
                        be drawn.
        ignore_kpt_idx: A list of the keypoint indecies that you would like to skip when drawing
                        the keypoints. The skeleton segments connected to ignored keypoints will
                        also be excluded.
        kpt_idx_offset: If your keypoint skeleton indecies do not start at 0, you need to set an
                        offset so that the index will start at 0.
                        Example: If your keypoint index starts at 1, use kpt_idx_offset=-1.
        skeleton_thickness: The thickness of the skeleton segments that are to be drawn.
        skeleton_color: The color of the skeleton segments that are to be drawn.
        show_bbox: If False, the bbox will not be drawn at all.
        show_kpt: If False, the keypoints will not be drawn at all.
        show_skeleton: If False, the keypoint skeleton will not be drawn at all.
        show_seg: If False, the segmentation will not be drawn at all.
        """
        coco_ann = self.annotations.get_obj_from_id(ann_id)
        result = img.copy()

        if len(coco_ann.keypoints) > 0:
            vis_keypoints_arr = coco_ann.keypoints.to_numpy(demarcation=True)[:, :2]
            kpt_visibility = coco_ann.keypoints.to_numpy(demarcation=True)[:, 2:].reshape(-1)
            base_ignore_kpt_idx = np.argwhere(np.array(kpt_visibility) == 0.0).reshape(-1).tolist()
            ignore_kpt_idx_list = ignore_kpt_idx + list(set(base_ignore_kpt_idx) - set(ignore_kpt_idx))
        else:
            vis_keypoints_arr = np.array([])
            kpt_visibility = np.array([])
            ignore_kpt_idx_list = []
        coco_cat = self.categories.get_obj_from_id(coco_ann.category_id)
        for draw_target in draw_order:
            if draw_target.lower() == 'bbox':
                if show_bbox:
                    result = draw_bbox(
                        img=result, bbox=coco_ann.bbox, color=bbox_color, thickness=bbox_thickness, text=coco_cat.name,
                        label_thickness=bbox_label_thickness, label_only=bbox_label_only
                    )
            elif draw_target.lower() == 'seg':
                if show_seg:
                    result = draw_segmentation(
                        img=result, segmentation=coco_ann.segmentation, color=seg_color, transparent=seg_transparent
                    )
            elif draw_target.lower() == 'kpt':
                if show_kpt:
                    result = draw_keypoints(
                        img=result, keypoints=vis_keypoints_arr,
                        radius=kpt_radius, color=kpt_color, keypoint_labels=coco_cat.keypoints,
                        show_keypoints_labels=show_kpt_labels, label_thickness=kpt_label_thickness,
                        label_only=kpt_label_only, ignore_kpt_idx=ignore_kpt_idx_list
                    )
            elif draw_target.lower() == 'skeleton':
                if show_skeleton:
                    result = draw_skeleton(
                        img=result, keypoints=vis_keypoints_arr,
                        keypoint_skeleton=coco_cat.skeleton, index_offset=kpt_idx_offset,
                        thickness=skeleton_thickness, color=skeleton_color, ignore_kpt_idx=ignore_kpt_idx_list
                    )
            else:
                logger.error(f'Invalid target: {draw_target}')
                logger.error(f"Valid targets: {['bbox', 'seg', 'kpt', 'skeleton']}")
                raise Exception
        return result

    def get_preview(
        self, image_id: int,
        draw_order: list=['seg', 'bbox', 'skeleton', 'kpt'],
        bbox_color: list=[0, 255, 255], bbox_thickness: list=2, # BBox
        bbox_show_label: bool=True, bbox_label_thickness: int=None,
        bbox_label_only: bool=False,
        seg_color: list=[255, 255, 0], seg_transparent: bool=True, # Segmentation
        kpt_radius: int=4, kpt_color: list=[0, 0, 255], # Keypoints
        show_kpt_labels: bool=True, kpt_label_thickness: int=1,
        kpt_label_only: bool=False, ignore_kpt_idx: list=[],
        kpt_idx_offset: int=0,
        skeleton_thickness: int=5, skeleton_color: list=[255, 0, 0], # Skeleton
        details_corner_pos_ratio: float=0.02, details_height_ratio: float=0.10, # Details
        details_leeway: float=0.4, details_color: list=[255, 0, 255],
        details_thickness: int=2,
        show_bbox: bool=True, show_kpt: bool=True, # Show Flags
        show_skeleton: bool=True, show_seg: bool=True,
        show_details: bool=False
    ) -> np.ndarray:
        """
        Returns a preview of the image in the dataset that corresponds to image_id.
        Annotations are included in the preview.

        image_id: The id of the image that you would like to preview.
        draw_order: The order in which you would like to draw (render) the annotations.
                    Example: If you specify 'bbox' after 'seg', the bounding box will be
                    drawn after the segmentation is drawn.
        bbox_color: The color of the bbox that is to be drawn.
        bbox_thickness: The thickness of the bbox that is to be drawn.
        bbox_show_label: If True, the label of the bbox will be drawn directly above it.
        bbox_label_thickness: The thickness of the bbox label in the event that it is drawn.
        bbox_label_only: If you would rather not draw the bounding box and only show the label,
                         set this to True.
        seg_color: The color of the segmentation that is to be drawn.
        seg_transparent: If True, the segmentation that is drawn will be transparent.
                         If False, the segmentation will be a solid color.
        kpt_radius: The radius of the keypoints that are to be drawn.
        kpt_color: The color of the keypoints that are to be drawn.
        show_kpt_labels: If True, the labels of the keypoints will be drawn directly above each
                         keypoint.
        kpt_label_thickness: The thickness of the keypoint labels in the event that they are drawn.
        kpt_label_only: If True, the keypoints will not be drawn and only the keypoint labels will
                        be drawn.
        ignore_kpt_idx: A list of the keypoint indecies that you would like to skip when drawing
                        the keypoints. The skeleton segments connected to ignored keypoints will
                        also be excluded.
        kpt_idx_offset: If your keypoint skeleton indecies do not start at 0, you need to set an
                        offset so that the index will start at 0.
                        Example: If your keypoint index starts at 1, use kpt_idx_offset=-1.
        skeleton_thickness: The thickness of the skeleton segments that are to be drawn.
        skeleton_color: The color of the skeleton segments that are to be drawn.
        details_corner_pos_ratio: Ratio that determines where the details are positioned with respect
                                  to the upper lefthand cornder.
        details_height_ratio: The height ratio of the details box.
        details_leeway: The leeway between lines in the details box.
        details_color: The color of the text in the details box.
        details_thickness: The thickness of the text in the details box.
        show_bbox: If False, the bbox will not be drawn at all.
        show_kpt: If False, the keypoints will not be drawn at all.
        show_skeleton: If False, the keypoint skeleton will not be drawn at all.
        show_seg: If False, the segmentation will not be drawn at all.
        show_details: If True, the filename of the current frame and other information will be written to the screen.
        """
        coco_image = self.images.get_obj_from_id(image_id)
        img = cv2.imread(coco_image.coco_url)
        for coco_ann in self.annotations.get_annotations_from_imgIds([coco_image.id]):
            img = self.draw_annotation(
                img=img, ann_id=coco_ann.id,
                draw_order=draw_order,
                bbox_color=bbox_color, bbox_thickness=bbox_thickness, # BBox
                bbox_show_label=bbox_show_label, bbox_label_thickness=bbox_label_thickness,
                bbox_label_only=bbox_label_only,
                seg_color=seg_color, seg_transparent=seg_transparent, # Segmentation
                kpt_radius=kpt_radius, kpt_color=kpt_color, # Keypoints
                show_kpt_labels=show_kpt_labels, kpt_label_thickness=kpt_label_thickness,
                kpt_label_only=kpt_label_only, ignore_kpt_idx=ignore_kpt_idx,
                kpt_idx_offset=kpt_idx_offset,
                skeleton_thickness=skeleton_thickness, skeleton_color=skeleton_color, # Skeleton
                details_corner_pos_ratio=details_corner_pos_ratio,
                details_height_ratio=details_height_ratio,
                details_leeway=details_leeway, details_color=details_color,
                details_thickness=details_thickness,
                show_bbox=show_bbox, show_kpt=show_kpt,
                show_skeleton=show_skeleton, show_seg=show_seg,
                show_details=show_details
            )
        if show_details:
            img_h, img_w = img.shape[:2]
            coco_anns = self.annotations.get_annotations_from_imgIds([coco_image.id])
            coco_ann_id_list = [coco_ann.id for coco_ann in coco_anns]
            coco_ann_id_list.sort()
            img = draw_text_rows_at_point(
                img=img,
                row_text_list=[
                    f'{coco_image.file_name}',
                    f'image_id: {coco_image.id}',
                    f'ann_ids: {coco_ann_id_list}'
                ],
                x=int(details_corner_pos_ratio*img_w), y=int(details_corner_pos_ratio*img_h),
                combined_row_height=int(details_height_ratio*img_h),
                leeway=details_leeway,
                color=details_color,
                thickness=details_thickness
            )
        return img

    def display_preview(
        self,
        start_idx: int=0, end_idx: int=None, preview_width: int=1000,
        draw_order: list=['seg', 'bbox', 'skeleton', 'kpt'],
        bbox_color: list=[0, 255, 255], bbox_thickness: list=2, # BBox
        bbox_show_label: bool=True, bbox_label_thickness: int=None,
        bbox_label_only: bool=False,
        seg_color: list=[255, 255, 0], seg_transparent: bool=True, # Segmentation
        kpt_radius: int=4, kpt_color: list=[0, 0, 255], # Keypoints
        show_kpt_labels: bool=True, kpt_label_thickness: int=1,
        kpt_label_only: bool=False, ignore_kpt_idx: list=[],
        kpt_idx_offset: int=0,
        skeleton_thickness: int=5, skeleton_color: list=[255, 0, 0], # Skeleton
        details_corner_pos_ratio: float=0.02, details_height_ratio: float=0.10, # Details
        details_leeway: float=0.4, details_color: list=[255, 0, 255],
        details_thickness: int=2,
        show_bbox: bool=True, show_kpt: bool=True, # Show Flags
        show_skeleton: bool=True, show_seg: bool=True,
        show_details: bool=False
    ):
        """
        Displays a preview of the dataset in a popup window.
        Annotations are included in the preview.

        start_idx: The index that you would like to start previewing the dataset at.
                   Default: 0
        end_idx: The index that you would like to stop previewing the dataset at.
                 Defualt: None
        preview_width: The width of the window that you would like to display the
                       preview in. Default: 1000
        draw_order: The order in which you would like to draw (render) the annotations.
                    Example: If you specify 'bbox' after 'seg', the bounding box will be
                    drawn after the segmentation is drawn.
        bbox_color: The color of the bbox that is to be drawn.
        bbox_thickness: The thickness of the bbox that is to be drawn.
        bbox_show_label: If True, the label of the bbox will be drawn directly above it.
        bbox_label_thickness: The thickness of the bbox label in the event that it is drawn.
        bbox_label_only: If you would rather not draw the bounding box and only show the label,
                         set this to True.
        seg_color: The color of the segmentation that is to be drawn.
        seg_transparent: If True, the segmentation that is drawn will be transparent.
                         If False, the segmentation will be a solid color.
        kpt_radius: The radius of the keypoints that are to be drawn.
        kpt_color: The color of the keypoints that are to be drawn.
        show_kpt_labels: If True, the labels of the keypoints will be drawn directly above each
                         keypoint.
        kpt_label_thickness: The thickness of the keypoint labels in the event that they are drawn.
        kpt_label_only: If True, the keypoints will not be drawn and only the keypoint labels will
                        be drawn.
        ignore_kpt_idx: A list of the keypoint indecies that you would like to skip when drawing
                        the keypoints. The skeleton segments connected to ignored keypoints will
                        also be excluded.
        kpt_idx_offset: If your keypoint skeleton indecies do not start at 0, you need to set an
                        offset so that the index will start at 0.
                        Example: If your keypoint index starts at 1, use kpt_idx_offset=-1.
        skeleton_thickness: The thickness of the skeleton segments that are to be drawn.
        skeleton_color: The color of the skeleton segments that are to be drawn.
        details_corner_pos_ratio: Ratio that determines where the details are positioned with respect
                                  to the upper lefthand cornder.
        details_height_ratio: The height ratio of the details box.
        details_leeway: The leeway between lines in the details box.
        details_color: The color of the text in the details box.
        details_thickness: The thickness of the text in the details box.
        show_bbox: If False, the bbox will not be drawn at all.
        show_kpt: If False, the keypoints will not be drawn at all.
        show_skeleton: If False, the keypoint skeleton will not be drawn at all.
        show_seg: If False, the segmentation will not be drawn at all.
        show_details: If True, the filename of the current frame and other information will be written to the screen.
        """
        last_idx = len(self.images) if end_idx is None else end_idx
        for coco_image in self.images[start_idx:last_idx]:
            img = self.get_preview(
                image_id=coco_image.id,
                draw_order=draw_order,
                bbox_color=bbox_color, bbox_thickness=bbox_thickness, # BBox
                bbox_show_label=bbox_show_label, bbox_label_thickness=bbox_label_thickness,
                bbox_label_only=bbox_label_only,
                seg_color=seg_color, seg_transparent=seg_transparent, # Segmentation
                kpt_radius=kpt_radius, kpt_color=kpt_color, # Keypoints
                show_kpt_labels=show_kpt_labels, kpt_label_thickness=kpt_label_thickness,
                kpt_label_only=kpt_label_only, ignore_kpt_idx=ignore_kpt_idx,
                kpt_idx_offset=kpt_idx_offset,
                details_corner_pos_ratio=details_corner_pos_ratio,
                details_height_ratio=details_height_ratio,
                details_leeway=details_leeway, details_color=details_color,
                details_thickness=details_thickness,
                skeleton_thickness=skeleton_thickness, skeleton_color=skeleton_color, # Skeleton
                show_bbox=show_bbox, show_kpt=show_kpt,
                show_skeleton=show_skeleton, show_seg=show_seg,
                show_details=show_details
            )
            quit_flag = cv_simple_image_viewer(img=img, preview_width=preview_width)
            if quit_flag:
                break

    def save_visualization(
        self, save_dir: str='vis_preview', show_preview: bool=False, preserve_filenames: bool=True,
        show_annotations: bool=True, overwrite: bool=False,
        start_idx: int=0, end_idx: int=None, preview_width: int=1000,
        draw_order: list=['seg', 'bbox', 'skeleton', 'kpt'],
        bbox_color: list=[0, 255, 255], bbox_thickness: list=2, # BBox
        bbox_show_label: bool=True, bbox_label_thickness: int=None,
        bbox_label_only: bool=False,
        seg_color: list=[255, 255, 0], seg_transparent: bool=True, # Segmentation
        kpt_radius: int=4, kpt_color: list=[0, 0, 255], # Keypoints
        show_kpt_labels: bool=True, kpt_label_thickness: int=1,
        kpt_label_only: bool=False, ignore_kpt_idx: list=[],
        kpt_idx_offset: int=0,
        skeleton_thickness: int=5, skeleton_color: list=[255, 0, 0], # Skeleton
        details_corner_pos_ratio: float=0.02, details_height_ratio: float=0.10, # Details
        details_leeway: float=0.4, details_color: list=[255, 0, 255],
        details_thickness: int=2,
        show_bbox: bool=True, show_kpt: bool=True, # Show Flags
        show_skeleton: bool=True, show_seg: bool=True,
        show_details: bool=False
    ):
        """
        Generates and saves visualizations of the annotations of this dataset to a dump folder.

        save_dir: The directory where you would like to save all of your dataset visualizations.
        show_preview: If True, the visualizations will be displayed in a popup window as they
                      are being generated.
        preserve_filenames: If true, the visualization images generated will use the same filename
                            as the file_name field in each COCO_Image.
                            Otherwise a unique filename will be automatically generated.
        show_annotations: If True, the annotations will be drawn on the visualizations.
                          Otherwise, no annotations will be drawn.
        overwrite: If True, the files contained in the save_dir will be deleted being generating
                   the visualizations.
        start_idx: The index that you would like to start previewing the dataset at.
                   Default: 0
        end_idx: The index that you would like to stop previewing the dataset at.
                 Defualt: None
        preview_width: The width of the window that you would like to display the
                       preview in. Default: 1000
        draw_order: The order in which you would like to draw (render) the annotations.
                    Example: If you specify 'bbox' after 'seg', the bounding box will be
                    drawn after the segmentation is drawn.
        bbox_color: The color of the bbox that is to be drawn.
        bbox_thickness: The thickness of the bbox that is to be drawn.
        bbox_show_label: If True, the label of the bbox will be drawn directly above it.
        bbox_label_thickness: The thickness of the bbox label in the event that it is drawn.
        bbox_label_only: If you would rather not draw the bounding box and only show the label,
                         set this to True.
        seg_color: The color of the segmentation that is to be drawn.
        seg_transparent: If True, the segmentation that is drawn will be transparent.
                         If False, the segmentation will be a solid color.
        kpt_radius: The radius of the keypoints that are to be drawn.
        kpt_color: The color of the keypoints that are to be drawn.
        show_kpt_labels: If True, the labels of the keypoints will be drawn directly above each
                         keypoint.
        kpt_label_thickness: The thickness of the keypoint labels in the event that they are drawn.
        kpt_label_only: If True, the keypoints will not be drawn and only the keypoint labels will
                        be drawn.
        ignore_kpt_idx: A list of the keypoint indecies that you would like to skip when drawing
                        the keypoints. The skeleton segments connected to ignored keypoints will
                        also be excluded.
        kpt_idx_offset: If your keypoint skeleton indecies do not start at 0, you need to set an
                        offset so that the index will start at 0.
                        Example: If your keypoint index starts at 1, use kpt_idx_offset=-1.
        skeleton_thickness: The thickness of the skeleton segments that are to be drawn.
        skeleton_color: The color of the skeleton segments that are to be drawn.
        details_corner_pos_ratio: Ratio that determines where the details are positioned with respect
                                  to the upper lefthand cornder.
        details_height_ratio: The height ratio of the details box.
        details_leeway: The leeway between lines in the details box.
        details_color: The color of the text in the details box.
        details_thickness: The thickness of the text in the details box.
        show_bbox: If False, the bbox will not be drawn at all.
        show_kpt: If False, the keypoints will not be drawn at all.
        show_skeleton: If False, the keypoint skeleton will not be drawn at all.
        show_seg: If False, the segmentation will not be drawn at all.
        show_details: If True, the filename of the current frame and other information will be written to the screen.
        """

        # Prepare save directory
        make_dir_if_not_exists(save_dir)
        if get_dir_contents_len(save_dir) > 0:
            if not overwrite:
                logger.error(f'save_dir={save_dir} is not empty.')
                logger.error(f"Hint: If you want to erase the directory's contents, use overwrite=True")
                raise Exception
            delete_all_files_in_dir(save_dir, ask_permission=False)

        if show_preview:
            # Prepare Viewer
            viewer = SimpleVideoViewer(preview_width=1000, window_name='Annotation Visualization')

        last_idx = len(self.images) if end_idx is None else end_idx
        total_iter = len(self.images[start_idx:last_idx])
        for coco_image in tqdm(self.images[start_idx:last_idx], total=total_iter, leave=False):
            if show_annotations:
                img = self.get_preview(
                    image_id=coco_image.id,
                    draw_order=draw_order,
                    bbox_color=bbox_color, bbox_thickness=bbox_thickness, # BBox
                    bbox_show_label=bbox_show_label, bbox_label_thickness=bbox_label_thickness,
                    bbox_label_only=bbox_label_only,
                    seg_color=seg_color, seg_transparent=seg_transparent, # Segmentation
                    kpt_radius=kpt_radius, kpt_color=kpt_color, # Keypoints
                    show_kpt_labels=show_kpt_labels, kpt_label_thickness=kpt_label_thickness,
                    kpt_label_only=kpt_label_only, ignore_kpt_idx=ignore_kpt_idx,
                    kpt_idx_offset=kpt_idx_offset,
                    skeleton_thickness=skeleton_thickness, skeleton_color=skeleton_color, # Skeleton
                    details_corner_pos_ratio=details_corner_pos_ratio,
                    details_height_ratio=details_height_ratio,
                    details_leeway=details_leeway, details_color=details_color,
                    details_thickness=details_thickness,
                    show_bbox=show_bbox, show_kpt=show_kpt,
                    show_skeleton=show_skeleton, show_seg=show_seg,
                    show_details=show_details
                )
            else:
                img = cv2.imread(coco_image.coco_url)

            if preserve_filenames:
                save_path = f'{save_dir}/{coco_image.file_name}'
                if file_exists(save_path):
                    logger.error(f"Your dataset contains multiple instances of the same filename.")
                    logger.error(f"Either make all filenames unique or use preserve_filenames=False")
                    raise Exception
                cv2.imwrite(save_path, img)
            else:
                file_extension = get_extension_from_filename(coco_image.file_name)
                save_path = get_next_dump_path(dump_dir=save_dir, file_extension=file_extension)
                cv2.imwrite(save_path, img)

            if show_preview:
                quit_flag = viewer.show(img)
                if quit_flag:
                    break

    def save_video(
        self, save_path: str='viz.mp4', show_preview: bool=False,
        fps: int=20, rescale_before_pad: bool=True,
        show_annotations: bool=True, overwrite: bool=False,
        start_idx: int=0, end_idx: int=None, preview_width: int=1000,
        draw_order: list=['seg', 'bbox', 'skeleton', 'kpt'],
        bbox_color: list=[0, 255, 255], bbox_thickness: list=2, # BBox
        bbox_show_label: bool=True, bbox_label_thickness: int=None,
        bbox_label_only: bool=False,
        seg_color: list=[255, 255, 0], seg_transparent: bool=True, # Segmentation
        kpt_radius: int=4, kpt_color: list=[0, 0, 255], # Keypoints
        show_kpt_labels: bool=True, kpt_label_thickness: int=1,
        kpt_label_only: bool=False, ignore_kpt_idx: list=[],
        kpt_idx_offset: int=0,
        skeleton_thickness: int=5, skeleton_color: list=[255, 0, 0], # Skeleton
        details_corner_pos_ratio: float=0.02, details_height_ratio: float=0.10, # Details
        details_leeway: float=0.4, details_color: list=[255, 0, 255],
        details_thickness: int=2,
        show_bbox: bool=True, show_kpt: bool=True, # Show Flags
        show_skeleton: bool=True, show_seg: bool=True,
        show_details: bool=False
    ):
        """
        save_path: Path to where you would like to save the visualization video of this dataset.
        show_preview: If True, the visualizations will be displayed in a popup window as they
                      are being generated.
        fps: The frames per second of the output video saved to save_path.
        rescale_before_pad: Since all output frames need to be of the same dimensions, output images
                            need to be rescaled and padded in order to fit the frame.
                            If rescale_before_pad=True, output images will be rescaled to fit with
                            either the vertical or horizontal borders of the frame before applying
                            padding.
                            If rescale_before_pad=False, the output images will only be padded.
        show_annotations: If True, the annotations will be drawn on the visualizations.
                          Otherwise, no annotations will be drawn.
        overwrite: If True, the files contained in the save_dir will be deleted being generating
                   the visualizations.
        start_idx: The index that you would like to start previewing the dataset at.
                   Default: 0
        end_idx: The index that you would like to stop previewing the dataset at.
                 Defualt: None
        preview_width: The width of the window that you would like to display the
                       preview in. Default: 1000
        draw_order: The order in which you would like to draw (render) the annotations.
                    Example: If you specify 'bbox' after 'seg', the bounding box will be
                    drawn after the segmentation is drawn.
        bbox_color: The color of the bbox that is to be drawn.
        bbox_thickness: The thickness of the bbox that is to be drawn.
        bbox_show_label: If True, the label of the bbox will be drawn directly above it.
        bbox_label_thickness: The thickness of the bbox label in the event that it is drawn.
        bbox_label_only: If you would rather not draw the bounding box and only show the label,
                         set this to True.
        seg_color: The color of the segmentation that is to be drawn.
        seg_transparent: If True, the segmentation that is drawn will be transparent.
                         If False, the segmentation will be a solid color.
        kpt_radius: The radius of the keypoints that are to be drawn.
        kpt_color: The color of the keypoints that are to be drawn.
        show_kpt_labels: If True, the labels of the keypoints will be drawn directly above each
                         keypoint.
        kpt_label_thickness: The thickness of the keypoint labels in the event that they are drawn.
        kpt_label_only: If True, the keypoints will not be drawn and only the keypoint labels will
                        be drawn.
        ignore_kpt_idx: A list of the keypoint indecies that you would like to skip when drawing
                        the keypoints. The skeleton segments connected to ignored keypoints will
                        also be excluded.
        kpt_idx_offset: If your keypoint skeleton indecies do not start at 0, you need to set an
                        offset so that the index will start at 0.
                        Example: If your keypoint index starts at 1, use kpt_idx_offset=-1.
        skeleton_thickness: The thickness of the skeleton segments that are to be drawn.
        skeleton_color: The color of the skeleton segments that are to be drawn.
        details_corner_pos_ratio: Ratio that determines where the details are positioned with respect
                                  to the upper lefthand cornder.
        details_height_ratio: The height ratio of the details box.
        details_leeway: The leeway between lines in the details box.
        details_color: The color of the text in the details box.
        details_thickness: The thickness of the text in the details box.
        show_bbox: If False, the bbox will not be drawn at all.
        show_kpt: If False, the keypoints will not be drawn at all.
        show_skeleton: If False, the keypoint skeleton will not be drawn at all.
        show_seg: If False, the segmentation will not be drawn at all.
        show_details: If True, the filename of the current frame and other information will be written to the screen.
        """
        # Check Output Path
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at {save_path}')
            raise Exception

        # Prepare Video Writer
        dim_list = np.array([[coco_image.height, coco_image.width] for coco_image in self.images])
        max_h, max_w = dim_list.max(axis=0).tolist()
        recorder = Recorder(output_path=save_path, output_dims=(max_w, max_h), fps=fps)

        if show_preview:
            # Prepare Viewer
            viewer = SimpleVideoViewer(preview_width=1000, window_name='Annotation Visualization')

        last_idx = len(self.images) if end_idx is None else end_idx
        total_iter = len(self.images[start_idx:last_idx])
        for coco_image in tqdm(self.images[start_idx:last_idx], total=total_iter, leave=False):
            if show_annotations:
                img = self.get_preview(
                    image_id=coco_image.id,
                    draw_order=draw_order,
                    bbox_color=bbox_color, bbox_thickness=bbox_thickness, # BBox
                    bbox_show_label=bbox_show_label, bbox_label_thickness=bbox_label_thickness,
                    bbox_label_only=bbox_label_only,
                    seg_color=seg_color, seg_transparent=seg_transparent, # Segmentation
                    kpt_radius=kpt_radius, kpt_color=kpt_color, # Keypoints
                    show_kpt_labels=show_kpt_labels, kpt_label_thickness=kpt_label_thickness,
                    kpt_label_only=kpt_label_only, ignore_kpt_idx=ignore_kpt_idx,
                    kpt_idx_offset=kpt_idx_offset,
                    skeleton_thickness=skeleton_thickness, skeleton_color=skeleton_color, # Skeleton
                    details_corner_pos_ratio=details_corner_pos_ratio,
                    details_height_ratio=details_height_ratio,
                    details_leeway=details_leeway, details_color=details_color,
                    details_thickness=details_thickness,
                    show_bbox=show_bbox, show_kpt=show_kpt,
                    show_skeleton=show_skeleton, show_seg=show_seg,
                    show_details=show_details
                )
            else:
                img = cv2.imread(coco_image.coco_url)

            if rescale_before_pad:
                img = scale_to_max(img=img, target_shape=[max_h, max_w])
            img = pad_to_max(img=img, target_shape=[max_h, max_w])
            recorder.write(img)

            if show_preview:
                quit_flag = viewer.show(img)
                if quit_flag:
                    break
        recorder.close()


from __future__ import annotations
import operator
from logger import logger
from common_utils.check_utils import check_type

class ID_Map:
    def __init__(self, unique_key: str, old_id: int, new_id: int):
        self.unique_key = unique_key
        self.old_id = old_id
        self.new_id = new_id

class ID_Mapper:
    def __init__(self):
        self.id_maps = []

    def __len__(self) -> int:
        return len(self.id_maps)

    def __getitem__(self, idx: int) -> ID_Mapper:
        if len(self.id_maps) == 0:
            logger.error(f"ID_Mapper is empty.")
            raise IndexError
        elif idx < 0 or idx >= len(self.id_maps):
            logger.error(f"Index out of range: {idx}")
            raise IndexError
        else:
            return self.id_maps[idx]

    def __setitem__(self, idx: int, value: ID_Mapper):
        check_type(value, valid_type_list=[ID_Mapper])
        self.id_maps[idx] = value

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> ID_Mapper:
        if self.n < len(self.id_maps):
            result = self.id_maps[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def sort(self):
        self.id_maps.sort(key=operator.attrgetter('old_id'), reverse=False)

    def add_id_map(self, id_map: ID_Map):
        self.id_maps.append(id_map)

    def add(self, unique_key: str, old_id: int, new_id: int):
        new_id_map = ID_Map(
            unique_key=unique_key,
            old_id=old_id,
            new_id=new_id
        )
        self.add_id_map(new_id_map)
    
    def get_new_id(self, unique_key: str, old_id: int) -> (bool, int):
        found = False
        new_id = None
        for id_map in self.id_maps:
            if id_map.unique_key == unique_key and id_map.old_id == old_id:
                found = True
                new_id = id_map.new_id
        return found, new_id

class COCO_Mapper_Handler:
    def __init__(self):
        self.license_mapper = ID_Mapper()
        self.image_mapper = ID_Mapper()
        self.annotation_mapper = ID_Mapper()
        self.category_mapper = ID_Mapper()


import cv2
import numpy as np
import operator
import matplotlib.pyplot as plt
import matplotlib.image as mpimage
import matplotlib.patches as patches
from pycocotools.coco import COCO
from logger import logger
from common_utils.file_utils import delete_dir_if_exists, make_dir, file_exists
from common_utils.path_utils import get_rootname_from_filename, get_extension_from_filename
from common_utils.cv_drawing_utils import draw_bbox, draw_segmentation, draw_skeleton, draw_keypoints
from common_utils.common_types.bbox import BBox
from common_utils.common_types.segmentation import Segmentation

from ..coco_annotation import COCO_AnnotationFileParser
from ...structs import COCO_Image, COCO_Annotation

class COCO_API_COCOAnnotationVisualizer:
    """
    Obsolete. Please use COCOAnnotationVisualizer instead.
    """
    def __init__(
        self, img_dir: str, coco_annotation_path: str, visualization_dump_dir: str, included_categories: list
    ):
        self.img_dir = img_dir
        self.coco_annotation_path = coco_annotation_path
        self.visualization_dump_dir = visualization_dump_dir
        self.included_categories = included_categories

    def save(self, coco: COCO, img: dict, anns: list, show_bbox: bool=False, filename_key: str='file_name'):
        I = mpimage.imread(f"{self.img_dir}/{img[filename_key]}")
        plt.axis('off')
        plt.imshow(I)
        ax = plt.gca()
        ax.set_axis_off()
        ax.margins(0,0)
        ax.xaxis.set_major_locator(plt.NullLocator())
        ax.yaxis.set_major_locator(plt.NullLocator())
        coco.showAnns(anns)

        if show_bbox:
            for ann in anns:
                xmin, ymin, width, height = ann['bbox']
                # Create a Rectangle patch
                rect = patches.Rectangle((xmin, ymin), width, height, linewidth=1, edgecolor='yellow', facecolor='none')
                # Add the patch to the Axes
                ax.add_patch(rect)

        rootname = str(img['id'])
        while len(rootname) < 6:
            rootname = '0' + rootname
        img_filename = f"{rootname}.jpg"
        plt.savefig(f'{self.visualization_dump_dir}/{img_filename}', bbox_inches='tight', pad_inches=0)
        logger.info(f'Created {self.visualization_dump_dir}/{img_filename}')
        plt.clf()

    def get_data(self, coco: COCO):
        catIds = coco.getCatIds(catNms=[self.included_categories])
        imgIds = coco.getImgIds(catIds=catIds)
        annIds = coco.getAnnIds()
        imgs = coco.loadImgs(imgIds)
        return catIds, annIds, imgs

    def generate_visualizations(self, show_bbox: bool=False, filename_key: str='file_name', limit: int=None):
        delete_dir_if_exists(self.visualization_dump_dir)
        make_dir(self.visualization_dump_dir)
        coco = COCO(annotation_file=self.coco_annotation_path)
        catIds, annIds, imgs = self.get_data(coco)

        for i, img in enumerate(imgs):
            if limit is not None and i >= limit:
                break
            annIds = coco.getAnnIds(imgIds=img['id'], catIds=catIds, iscrowd=None)
            anns = coco.loadAnns(annIds)

            has_keypoints = False
            for ann in anns:
                if ann['num_keypoints'] > 0:
                    has_keypoints = True
                    break
            if not has_keypoints and not show_bbox:
                continue

            self.save(coco, img, anns, show_bbox=show_bbox, filename_key=filename_key)

class COCOAnnotationVisualizer:
    """
    Visualize COCO Annotations given the annotation json path and corresponding image directory.
    """
    def __init__(
        self, img_dir: str, coco_annotation_path: str, visualization_dump_dir: str, included_categories: list,
        bbox_color: list=[0, 255, 255], kpt_color: list=[0, 0, 255], skeleton_color: list=[255, 0, 0], seg_color: list=[255, 255, 0],
        transparent_seg: bool=True, kpt_radius: int=10, bbox_thickness: int=2, skeleton_thickness: int=2, show_cat_name: bool=False,
        skeleton_index_offset: int=0,
        show_bbox: bool=True, show_kpts: bool=True, show_skeleton: bool=True, show_seg: bool=True,
        show_bbox_labels: bool=False, show_kpt_labels: bool=False, bbox_label_thickness: int=None, kpt_label_thickness: int=None,
        show_bbox_labels_only: bool=False, show_kpt_labels_only: bool=False,
        render_order: list=['bbox', 'seg', 'skeleton', 'kpt'],
        viz_limit: int=None
    ):
        self.img_dir = img_dir
        self.coco_annotation_path = coco_annotation_path
        self.visualization_dump_dir = visualization_dump_dir
        self.included_categories = included_categories

        self.parser = COCO_AnnotationFileParser(coco_annotation_path)
        self.parser.load()

        self.bbox_color = bbox_color
        self.kpt_color = kpt_color
        self.skeleton_color = skeleton_color
        self.seg_color = seg_color

        self.transparent_seg = transparent_seg
        self.kpt_radius = kpt_radius
        self.bbox_thickness = bbox_thickness
        self.skeleton_thickness = skeleton_thickness
        self.show_cat_name = show_cat_name
        self.skeleton_index_offset = skeleton_index_offset

        self.show_bbox = show_bbox
        self.show_kpts = show_kpts
        self.show_skeleton = show_skeleton
        self.show_seg = show_seg

        self.show_bbox_labels = show_bbox_labels
        self.show_kpt_labels = show_kpt_labels
        self.bbox_label_thickness = bbox_label_thickness
        self.kpt_label_thickness = kpt_label_thickness
        self.show_bbox_labels_only = show_bbox_labels_only
        self.show_kpt_labels_only = show_kpt_labels_only

        self.render_order = render_order

        self.viz_limit = viz_limit
        self.viz_count = 0

    def _draw(self, img: np.ndarray, bbox: BBox, kpt_list: list, kpt_skeleton: list, kpt_label_list: list, seg: Segmentation, cat_name: str) -> np.ndarray:
        for render_target in self.render_order:
            if render_target.lower() in ['seg', 'segmentation']:
                if self.show_seg:
                    img = draw_segmentation(img=img, segmentation=seg, color=self.seg_color, transparent=self.transparent_seg)
            elif render_target.lower() in ['skel', 'skeleton']:
                if self.show_skeleton:
                    img = draw_skeleton(
                        img=img, keypoints=kpt_list, keypoint_skeleton=kpt_skeleton, index_offset=self.skeleton_index_offset,
                        thickness=self.skeleton_thickness, color=self.skeleton_color
                    )
            elif render_target.lower() in ['kpt', 'kpts', 'keypoint', 'keypoints']:
                if self.show_kpts:
                    if self.kpt_label_thickness is not None:
                        kpt_label_thickness = self.kpt_label_thickness
                    else:
                        kpt_label_thickness = self.kpt_radius - 1 if self.kpt_radius > 1 else 1
                    img = draw_keypoints(
                        img=img, keypoints=kpt_list, radius=self.kpt_radius, color=self.kpt_color, keypoint_labels=kpt_label_list,
                        show_keypoints_labels=self.show_kpt_labels, label_thickness=kpt_label_thickness, label_only=self.show_kpt_labels_only
                    )
            elif render_target.lower() in ['bbox', 'bounding_box', 'bounding box']:
                if self.show_bbox:
                    bbox_text = cat_name if self.show_bbox_labels or self.show_bbox_labels_only else None
                    bbox_label_thickness = self.bbox_label_thickness if self.bbox_label_thickness is not None else self.bbox_thickness
                    img = draw_bbox(
                        img=img, bbox=bbox, color=self.bbox_color, thickness=self.bbox_thickness,
                        text=bbox_text, label_thickness=bbox_label_thickness, label_only=self.show_bbox_labels_only
                    )
            else:
                logger.error(f"Invalid render_target: {render_target}")
                logger.error(f"Options: seg, skeleton, kpt, bbox")
                raise Exception
        return img

    def save(
        self, img: np.ndarray,
        bbox_list: list,
        kpt_list_list: list, kpt_skeleton_list: list, kpt_label_list_list: list,
        seg_list: list, cat_name_list: list,
        file_name: str
    ):
        result = img.copy()
        save_path = f"{self.visualization_dump_dir}/{file_name}"
        retry_count = 0
        while file_exists(save_path):
            rootname = get_rootname_from_filename(file_name)
            extension = get_extension_from_filename(file_name)
            save_path = f"{self.visualization_dump_dir}/{rootname}_{retry_count}.{extension}"
            if retry_count == 9:
                logger.error(f"Can't resolve save_path.")
                raise Exception
            retry_count += 1

        for bbox, kpt_list, kpt_skeleton, kpt_label_list, seg, cat_name in \
            zip(bbox_list, kpt_list_list, kpt_skeleton_list, kpt_label_list_list, seg_list, cat_name_list):
            result = self._draw(
                img=result,
                bbox=bbox, kpt_list=kpt_list,
                kpt_skeleton=kpt_skeleton, kpt_label_list=kpt_label_list,
                seg=seg, cat_name=cat_name
            )
        cv2.imwrite(filename=save_path, img=result)
        logger.info(f"Wrote {save_path}")
        self.viz_count += 1

    def generate_visualizations(self, do_sort: bool=False):
        delete_dir_if_exists(self.visualization_dump_dir)
        make_dir(self.visualization_dump_dir)

        if do_sort:
            self.parser.images.image_list.sort(key=operator.attrgetter('file_name'))

        for coco_image in self.parser.images.image_list:
            coco_image = COCO_Image.buffer(coco_image)
            img_path = f"{self.img_dir}/{coco_image.file_name}"
            img = cv2.imread(img_path)
            coco_annotation_list = self.parser.annotations.get_annotations_from_imgIds(imgIds=[coco_image.id])
            bbox_list = []
            kpt_list_list = []
            kpt_skeleton_list = []
            kpt_label_list_list = []
            seg_list = []
            cat_name_list = []
            for coco_annotation in coco_annotation_list:
                coco_annotation = COCO_Annotation.buffer(coco_annotation)
                bbox = coco_annotation.bounding_box
                keypoint_list = coco_annotation.get_keypoint_list()
                kpt_list = [[kpt.x, kpt.y] for kpt in keypoint_list]
                seg = Segmentation.from_list(points_list=coco_annotation.segmentation)
                bbox_list.append(bbox)
                kpt_list_list.append(kpt_list)
                seg_list.append(seg)
                coco_category = self.parser.categories.get_category_from_id(id=coco_annotation.category_id)
                cat_name_list.append(coco_category.name)
                kpt_skeleton_list.append(coco_category.skeleton)
                kpt_label_list_list.append(coco_category.keypoints)
            self.save(
                img=img,
                bbox_list=bbox_list,
                kpt_list_list=kpt_list_list,
                kpt_skeleton_list=kpt_skeleton_list,
                kpt_label_list_list=kpt_label_list_list,
                seg_list=seg_list,
                cat_name_list=cat_name_list,
                file_name=coco_image.file_name
            )
            if self.viz_limit is not None and self.viz_count >= self.viz_limit:
                break


import numpy as np
from logger import logger
from ...labelme.annotation import Shape
from ..checks import check_shape_type
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
from ..common_types import BBox

def polygon2segmentation(polygon: Shape):
    check_shape_type(shape=polygon, shape_type='polygon')
    segmentation = []
    for point in polygon.points:
        segmentation.append(round(point[0], 2))
        segmentation.append(round(point[1], 2))
    return segmentation

def polygon2bbox(polygon: Shape) -> BBox:
    check_shape_type(shape=polygon, shape_type='polygon')
    x_values = np.array(polygon.points).T[0]
    y_values = np.array(polygon.points).T[1]
    xmin = round(x_values.min(), 2)
    ymin = round(y_values.min(), 2)
    xmax = round(x_values.max(), 2)
    ymax = round(y_values.max(), 2)
    return BBox(x=xmin, y=ymin, width=xmax-xmin, height=ymax-ymin)

def point_inside_polygon(point: Shape, polygon: Shape):
    check_shape_type(shape=point, shape_type='point')
    check_shape_type(shape=polygon, shape_type='polygon')
    shapely_point = Point(point.points[0])
    shapely_polygon = Polygon(polygon.points)
    return shapely_point.within(shapely_polygon)

def int_skeleton2str_skeleton(int_skeleton: list, keypoint_labels: list):
    str_skeleton = []
    for int_bone in int_skeleton:
        bone_start = keypoint_labels[int_bone[0]-1] 
        bone_end = keypoint_labels[int_bone[1]-1]
        str_bone = [bone_start, bone_end]
        str_skeleton.append(str_bone)
    return str_skeleton

def labeled_points2keypoints(keypoint_labels: list, labeled_points: list, img_path: str, annotation_path: str) -> list:
    """
    Note that keypoint entries are of the format (x, y, v), where
    v = 0: not labeled (in which case x=y=0)
    v = 1: labeled but not visible
    v = 2: labeled and visible
    """
    keypoint_registry = {}
    for keypoint_label in keypoint_labels:
        keypoint_registry[keypoint_label] = None
    
    for point in labeled_points:
        if point.label not in keypoint_labels:
            logger.error(f"Point label {point.label} not found in keypoint labels.")
            logger.error(f"Keypoint Labels: {keypoint_labels}")
            logger.error(f"Please modify the following image/annotation pair:")
            logger.error(f"\tImage: {img_path}")
            logger.error(f"\tAnnotation: {annotation_path}")
            raise Exception
        if keypoint_registry[point.label] is not None:
            logger.error(f"{point.label} point {point.points[0]} already exists in keypoint_registry.")
            logger.error(f"keypoint_registry[{point.label}]: {keypoint_registry[point.label].points[0]}")
            logger.error(f"Please modify the following image/annotation pair:")
            logger.error(f"\tImage: {img_path}")
            logger.error(f"\tAnnotation: {annotation_path}")
            raise Exception
        keypoint_registry[point.label] = point

    keypoints = []
    for point in keypoint_registry.values():
        if point is None:
            entry = [0, 0, 0] # not labeled
            keypoints.extend(entry)
        elif point.hidden:
            entry = [point.points[0][0], point.points[0][1], 1] # labeled but not visible
            keypoints.extend(entry)
        else:
            entry = [point.points[0][0], point.points[0][1], 2] # labeled and visible
            keypoints.extend(entry)
    return keypoints

def rectangle2bbox(rectangle: Shape) -> BBox:
    check_shape_type(shape=rectangle, shape_type='rectangle')
    p0, p1 = rectangle.points
    xmin = round(min(p0[0], p1[0]), 2)
    xmax = round(max(p0[0], p1[0]), 2)
    ymin = round(min(p0[1], p1[1]), 2)
    ymax = round(max(p0[1], p1[1]), 2)
    return BBox(x=xmin, y=ymin, width=xmax-xmin, height=ymax-ymin)

def point_inside_rectangle(point: Shape, rectangle: Shape):
    check_shape_type(shape=point, shape_type='point')
    check_shape_type(shape=rectangle, shape_type='rectangle')
    px = point.points[0][0]
    py = point.points[0][1]
    bbox = rectangle2bbox(rectangle)
    if px >= bbox.xmin and px <= bbox.xmax and py >= bbox.ymin and py <= bbox.ymax:
        return True
    else:
        return False


from __future__ import annotations
from typing import List
from logger import logger
from common_utils.path_utils import rel_to_abs_path
from common_utils.check_utils import check_type, check_type_from_list

class Path:
    def __init__(self, path_str: str):
        self.path_str = path_str

    def __str__(self) -> str:
        return self.path_str

    def __repr__(self) -> str:
        return self.__str__()

    def __key(self) -> tuple:
        return tuple([self.__class__] + list(self.__dict__.values()))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self.__key() == other.__key()
        return NotImplemented

    def __len__(self) -> int:
        return len(self.split())

    def __getitem__(self, idx: int) -> Path:
        if type(idx) is int:
            if len(self) == 0:
                logger.error(f"{type(self).__name__} is empty.")
                raise IndexError
            elif idx >= len(self) or idx < -len(self):
                logger.error(f"Index out of range: {idx}")
                raise IndexError
            else:
                return Path(self.split()[idx])
        elif type(idx) is slice:
            return Path.from_split(self.split()[idx.start:idx.stop:idx.step])
        else:
            logger.error(f'Expected int or slice. Got type(idx)={type(idx)}')
            raise TypeError

    def __setitem__(self, idx: int, value: Path):
        check_type(value, valid_type_list=[Path, str])
        if type(value) is str:
            value0 = Path(value)
        else:
            value0 = value
        path_parts = self.split()
        if type(idx) is int:
            path_parts[idx] = value.path_str
        elif type(idx) is slice:
            path_parts[idx.start:idx.stop:idx.step] = value.split()
        else:
            logger.error(f'Expected int or slice. Got type(idx)={type(idx)}')
            raise TypeError

    def __delitem__(self, idx):
        if type(idx) is int:
            if len(self) == 0:
                logger.error(f"{type(self).__name__} is empty.")
                raise IndexError
            elif idx >= len(self) or idx < -len(self):
                logger.error(f"Index out of range: {idx}")
                raise IndexError
            else:
                path_parts = self.split()
                del path_parts[idx]
                result = Path.from_split(path_parts)
                self.path_str = result.to_str()
        elif type(idx) is slice:
            path_parts = self.split()
            del path_parts[idx.start:idx.stop:idx.step]
            result = Path.from_split(path_parts)
            self.path_str = result.to_str()
        else:
            logger.error(f'Expected int or slice. Got type(idx)={type(idx)}')
            raise TypeError

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> Path:
        if self.n < len(self):
            result = Path(self.split()[self.n])
            self.n += 1
            return result
        else:
            raise StopIteration

    def __add__(self, other: Path) -> Path:
        return Path.from_split(self.split() + other.split()).prune_slashes()

    @classmethod
    def buffer(cls, path: Path) -> Path:
        return path

    def copy(self) -> Path:
        return Path(self.path_str)

    def split(self) -> List[str]:
        result = self.path_str.split('/')
        result = [result_part for result_part in result]
        return result

    @classmethod
    def from_split(self, str_list: List[str]) -> Path:
        return Path('/'.join(str_list))

    def head(self) -> Path:
        return self[0]

    def no_head(self) -> Path:
        return self[1:]

    def tail(self) -> Path:
        return self[-1]

    def no_tail(self) -> Path:
        return self[:-1]

    def to_str(self) -> str:
        return self.path_str

    def pop_head(self) -> Path:
        result = self.head()
        self.path_str = self.no_head().to_str()
        return result

    def pop_tail(self) -> Path:
        result = self.tail()
        self.path_str = self.no_tail().to_str()
        return result

    def push_head(self, path_part: Path):
        check_type(path_part, valid_type_list=[Path, str])
        if type(path_part) is str:
            path_part0 = Path(path_part)
        else:
            path_part0 = path_part
        return Path.from_split([path_part0.path_str] + self.split())

    def push_tail(self, path_part: Path):
        check_type(path_part, valid_type_list=[Path, str])
        if type(path_part) is str:
            path_part0 = Path(path_part)
        else:
            path_part0 = path_part
        return Path.from_split(self.split() + [path_part0.path_str])

    def get_extension(self) -> str:
        if '.' in self.tail().path_str:
            return self.tail().path_str.split('.')[-1]
        else:
            return ''

    def has_extension(self) -> bool:
        return self.get_extension() != ''

    def abs(self) -> Path:
        return Path(rel_to_abs_path(self.to_str()))

    def prune_slashes(self) -> Path:
        path_str = self.to_str()
        while '//' in path_str:
            path_str = path_str.replace('//', '/')
        return Path(path_str)

    @classmethod
    def get_unique_paths(cls, paths: List[Path]) -> List[Path]:
        path_str_list = [path.path_str for path in paths]
        unique_path_str_list = list(dict.fromkeys(path_str_list))
        unique_paths = [Path(unique_path_str) for unique_path_str in unique_path_str_list]
        return unique_paths

    @classmethod
    def has_common_head(cls, paths: List[Path]) -> bool:
        if len(paths) > 1:
            path_len_list = [len(path) for path in paths]
            if 0 in path_len_list:
                return False
            path_heads = [path.head() for path in paths]
            unique_path_heads = cls.get_unique_paths(path_heads)
            if len(unique_path_heads) == 1:
                return True
            else:
                return False
        elif len(paths) == 1 and len(paths[0]) > 0 and paths[0] != Path(''):
            return True
        else:
            return False

    @classmethod
    def has_common_tail(cls, paths: List[Path]) -> bool:
        if len(paths) > 1:
            path_len_list = [len(path) for path in paths]
            if 0 in path_len_list:
                return False
            path_tails = [path.tail() for path in paths]
            unique_path_tails = cls.get_unique_paths(path_tails)
            if len(unique_path_tails) == 1:
                return True
            else:
                return False
        elif len(paths) == 1 and len(paths[0]) > 0 and paths[0] != Path(''):
            return True
        else:
            return False

    @classmethod
    def get_common_head(cls, paths: List[Path]) -> (bool, Path):
        if len(paths) > 1:
            path_len_list = [len(path) for path in paths]
            if 0 in path_len_list:
                return False, None
            path_heads = [path.head() for path in paths]
            unique_path_heads = cls.get_unique_paths(path_heads)
            if len(unique_path_heads) == 1 and unique_path_heads[0] != Path(''):
                return True, unique_path_heads[0]
            else:
                return False, None
        elif len(paths) == 1 and len(paths[0]) > 0 and paths[0].head() != Path(''):
            return True, paths[0].head()
        else:
            return False, None

    @classmethod
    def get_common_tail(cls, paths: List[Path]) -> (bool, Path):
        if len(paths) > 1:
            path_len_list = [len(path) for path in paths]
            if 0 in path_len_list:
                return False, None
            path_tails = [path.tail() for path in paths]
            unique_path_tails = cls.get_unique_paths(path_tails)
            if len(unique_path_tails) == 1 and unique_path_tails[0] != Path(''):
                return True, unique_path_tails[0]
            else:
                return False, None
        elif len(paths) == 1 and len(paths[0]) > 0 and paths[0].tail() != Path(''):
            return True, paths[0].tail()
        else:
            return False, None

    def replace(self, old: Path, new: Path) -> Path:
        check_type_from_list([old, new], valid_type_list=[Path, str])
        if type(old) is str:
            old_path = Path(old)
        else:
            old_path = old
        if type(new) is str:
            new_path = Path(new)
        else:
            new_path = new
        result = self.to_str()
        result = result.replace(old_path.to_str(), new_path.to_str())
        while '//' in result:
            result = result.replace('//', '/')
        return Path(result)

    def possible_rel_paths(self) -> List[Path]:
        return [Path.from_split(self.split()[i:]) for i in range(len(self))]

    def possible_container_dirs(self) -> List[Path]:
        result = [
            Path.from_split(self.split()[:i]) if self.has_extension() else Path.from_split(self.split()[:i+1]) \
                for i in range(len(self))
        ]
        result = [path for path in result if path.path_str != '']
        return result

    @classmethod
    def del_duplicates(cls, path_list: List[Path]) -> List[Path]:
        path_str_list = list(dict.fromkeys([path.path_str for path in path_list]))
        return [Path(path_str) for path_str in path_str_list]

    @classmethod
    def get_common_container_dirs(cls, path_list: List[Path]) -> List[Path]:
        if len(path_list) == 0:
            logger.error(f"Encountered len(path_list) == 0")
            raise Exception
        possible_container_dirs_sets = []
        for path in path_list:
            possible_container_dirs_sets.append(set([path0.to_str() for path0 in path.possible_container_dirs()]))
        common_container_dirs = list(set.intersection(*possible_container_dirs_sets))
        return [Path(common_container_dir) for common_container_dir in common_container_dirs]

    @classmethod
    def get_longest_container_dir(cls, path_list: List[Path]) -> Path:
        if len(path_list) == 0:
            logger.error(f"Encountered len(path_list) == 0")
            raise Exception
        common_container_dirs = Path.get_common_container_dirs(path_list)
        longest_container_dir = None
        for common_container_dir in common_container_dirs:
            if longest_container_dir is None or len(common_container_dir) > len(longest_container_dir):
                longest_container_dir = common_container_dir
        return longest_container_dir

    @staticmethod
    def _root_src_tail2dst_head(src_path: Path, dst_path: Path) -> bool:
        check_type(src_path, valid_type_list=[Path])
        check_type(dst_path, valid_type_list=[Path])
        if len(src_path) > 0:
            tail = src_path.pop_tail()
            if tail != Path(''):
                dst_path.path_str = (tail + dst_path).to_str()
                success = True
            else:
                success = False
        else:
            success = False
        return success

    @classmethod
    def _root_src_head2dst_tail(cls, src_path: Path, dst_path: Path) -> bool:
        check_type(src_path, valid_type_list=[Path])
        check_type(dst_path, valid_type_list=[Path])
        if len(src_path) > 0:
            head = src_path.pop_head()
            if head != Path(''):
                dst_path.path_str = (dst_path + head).to_str()
                success = True
            else:
                success = False
        else:
            success = False
        return success

    @classmethod
    def tail2head(cls, src_obj: List[Path], dst_obj: List[Path]) -> bool:
        # TODO: Assertion test
        check_type_from_list([src_obj, dst_obj], valid_type_list=[Path, list])
        if type(src_obj) is list:
            check_type_from_list(src_obj, valid_type_list=[Path])
        if type(dst_obj) is list:
            check_type_from_list(dst_obj, valid_type_list=[Path])
        
        if type(src_obj) is Path:
            if type(dst_obj) is Path:
                success = cls._root_src_tail2dst_head(src_obj, dst_obj)
            elif type(dst_obj) is list:
                if len(src_obj) > 0:
                    tail = src_obj.pop_tail()
                    for dst_path in dst_obj:
                        dst_path.path_str = (tail + dst_path).to_str()
                    success = True
                else:
                    success = False
            else:
                raise Exception
        elif type(src_obj) is list:
            has_common_tail, common_tail = cls.get_common_tail(src_obj)
            if has_common_tail:
                if type(dst_obj) is Path:
                    dst_obj.path_str = (common_tail + dst_obj).to_str()
                    for src_path in src_obj:
                        src_path.path_str = src_path.no_tail().to_str()
                    success = True
                elif type(dst_obj) is list:
                    if len(src_obj) != len(dst_obj):
                        logger.error(f'len(src_obj) == {len(src_obj)} != {len(dst_obj)} == len(dst_obj)')
                        raise Exception
                    for dst_path in dst_obj:
                        dst_path.path_str = (common_tail + dst_path).to_str()
                    for src_path in src_obj:
                        src_path.path_str = src_path.no_tail().to_str()
                    success = True
                else:
                    raise Exception
            else:
                success = False
        else:
            raise Exception
        return success

    @classmethod
    def head2tail(cls, src_obj: List[Path], dst_obj: List[Path]) -> bool:
        check_type_from_list([src_obj, dst_obj], valid_type_list=[Path, list])
        if type(src_obj) is list:
            check_type_from_list(src_obj, valid_type_list=[Path])
        if type(dst_obj) is list:
            check_type_from_list(dst_obj, valid_type_list=[Path])
        
        if type(src_obj) is Path:
            if type(dst_obj) is Path:
                success = cls._root_src_head2dst_tail(src_obj, dst_obj)
            elif type(dst_obj) is list:
                if len(src_obj) > 0:
                    head = src_obj.pop_head()
                    for dst_path in dst_obj:
                        dst_path.path_str = (dst_path + head).to_str()
                    success = True
                else:
                    success = False
            else:
                raise Exception
        elif type(src_obj) is list:
            has_common_head, common_head = cls.get_common_head(src_obj)
            if has_common_head:
                if type(dst_obj) is Path:
                    dst_obj.path_str = (dst_obj + common_head).to_str()
                    for src_path in src_obj:
                        src_path.path_str = src_path.no_head().to_str()
                    success = True
                elif type(dst_obj) is list:
                    if len(src_obj) != len(dst_obj):
                        logger.error(f'len(src_obj) == {len(src_obj)} != {len(dst_obj)} == len(dst_obj)')
                        raise Exception
                    for dst_path in dst_obj:
                        dst_path.path_str = (dst_path + common_head).to_str()
                    for src_path in src_obj:
                        src_path.path_str = src_path.no_head().to_str()
                    success = True
                else:
                    raise Exception
            else:
                success = False
        else:
            raise Exception
        return success


from __future__ import annotations
import json
from logger import logger
from ..structs import COCO_Info, COCO_License_Handler, \
    COCO_Image_Handler, COCO_Annotation_Handler, COCO_Category_Handler
from ...util.utils.coco import COCO_Field_Buffer

class COCO_Writer:
    def __init__(
        self, info: COCO_Info, licenses: COCO_License_Handler,
        images: COCO_Image_Handler, annotations: COCO_Annotation_Handler,
        categories: COCO_Category_Handler, output_path: str
    ):
        self.info = info
        self.licenses = licenses
        self.images = images
        self.annotations = annotations
        self.categories = categories

        self.output_path = output_path

    @classmethod
    def from_buffer(self, buffer: COCO_Field_Buffer, output_path: str) -> COCO_Writer:
        return COCO_Writer(
            info=buffer.info,
            licenses=buffer.licenses,
            images=buffer.images,
            annotations=buffer.annotations,
            categories=buffer.categories,
            output_path=output_path
        )

    def build_json_dict(self, verbose: bool=False) -> dict:
        info_dict = self.get_info_dict()
        licenses_list = self.get_licenses_list()
        images_list = self.get_images_list()
        annotations_list = self.get_annotations_list()
        categories_list = self.get_categories_list()

        json_dict = {}
        json_dict['info'] = info_dict
        json_dict['licenses'] = licenses_list
        json_dict['images'] = images_list
        json_dict['annotations'] = annotations_list
        json_dict['categories'] = categories_list
        if verbose: logger.info("COCO json has been built successfully.")
        return json_dict
    
    def get_info_dict(self) -> dict:
        info_dict = {}
        info_dict['description'] = self.info.description
        info_dict['url'] = self.info.url
        info_dict['version'] = self.info.version
        info_dict['year'] = self.info.year
        info_dict['contributor'] = self.info.contributor
        info_dict['date_created'] = self.info.date_created
        return info_dict

    def get_licenses_list(self) -> list:
        licenses_list = []
        for coco_license in self.licenses.license_list:
            license_dict = {}
            license_dict['url'] = coco_license.url
            license_dict['id'] = coco_license.id
            license_dict['name'] = coco_license.name
            licenses_list.append(license_dict)
        return licenses_list

    def get_images_list(self) -> list:
        images_list = []
        for coco_image in self.images.image_list:
            image_dict = {}
            image_dict['license'] = coco_image.license_id
            image_dict['file_name'] = coco_image.file_name
            image_dict['coco_url'] = coco_image.coco_url
            image_dict['height'] = coco_image.height
            image_dict['width'] = coco_image.width
            image_dict['date_captured'] = coco_image.date_captured
            image_dict['flickr_url'] = coco_image.flickr_url
            image_dict['id'] = coco_image.id
            images_list.append(image_dict)
        return images_list

    def get_annotations_list(self) -> list:
        annotations_list = []
        for coco_annotation in self.annotations.annotation_list:
            annotation_dict = {}
            annotation_dict['segmentation'] = coco_annotation.segmentation
            annotation_dict['num_keypoints'] = coco_annotation.num_keypoints
            annotation_dict['area'] = coco_annotation.area
            annotation_dict['iscrowd'] = coco_annotation.iscrowd
            annotation_dict['keypoints'] = coco_annotation.keypoints
            annotation_dict['image_id'] = coco_annotation.image_id
            annotation_dict['bbox'] = coco_annotation.bbox
            annotation_dict['category_id'] = coco_annotation.category_id
            annotation_dict['id'] = coco_annotation.id
            annotations_list.append(annotation_dict)
        return annotations_list

    def get_categories_list(self) -> list:
        categories_list = []
        for coco_category in self.categories.category_list:
            category_dict = {}
            category_dict['supercategory'] = coco_category.supercategory
            category_dict['id'] = coco_category.id
            category_dict['name'] = coco_category.name
            category_dict['keypoints'] = coco_category.keypoints
            category_dict['skeleton'] = coco_category.skeleton
            categories_list.append(category_dict)
        return categories_list

    def write_json_dict(self, json_dict: dict, verbose: bool=False):
        json.dump(json_dict, open(self.output_path, 'w'), indent=2, ensure_ascii=False)
        if verbose: logger.info(f"JSON dict has been written to:\n{self.output_path}")

    def test(self):
        json_dict = self.build_json_dict()
        self.write_json_dict(json_dict)


from logger import logger
from common_utils.time_utils import get_present_year, get_present_time_Ymd
from common_utils.user_utils import get_username

from ..annotation import COCO_AnnotationFileParser
from ..structs import COCO_Info, COCO_License_Handler, COCO_Image_Handler, \
    COCO_Annotation_Handler, COCO_Category_Handler
from ..writer import COCO_Writer
from ...util.utils.coco import COCO_Mapper_Handler, COCO_Field_Buffer

def print_id_update(added_new: bool, old_id: int, new_id: int):
    header = 'ADDED' if added_new else 'LINKED'
    logger.info(f"{header}: {old_id} -> {new_id}")

class COCO_Annotations_Combiner:
    def __init__(self, img_dir_list: list, ann_path_list: list):
        # Constructor Parameters
        self.img_dir_list = img_dir_list
        self.ann_path_list = ann_path_list

        # Working Parameters
        self.buffer = COCO_Field_Buffer.from_scratch(
            description="COCO annotation generated by combining several coco annotations.",
            url='https://github.com/cm107/KeypointPose',
            version='1.0'
        )
        self.id_mapper = COCO_Mapper_Handler()

    def add_licenses(self, ann_key: str, coco_license_handler: COCO_License_Handler, verbose: bool=False):
        for coco_license in coco_license_handler.license_list:
            added_new, old_id, new_id = self.buffer.process_license(coco_license=coco_license, id_mapper=self.id_mapper, unique_key=ann_key)
            if verbose: print_id_update(added_new=added_new, old_id=old_id, new_id=new_id)

    def add_images(self, ann_key: str, coco_image_handler: COCO_Image_Handler, img_dir: str, update_img_paths: bool=True, verbose: bool=False):
        for coco_image in coco_image_handler.image_list:
            added_new, old_id, new_id = self.buffer.process_image(
                coco_image=coco_image, id_mapper=self.id_mapper, unique_key=ann_key, img_dir=img_dir, update_img_path=update_img_paths
            )
            if verbose: print_id_update(added_new=added_new, old_id=old_id, new_id=new_id)

    def add_annotations(self, ann_key: str, coco_annotation_handler: COCO_Annotation_Handler, verbose: bool=False):
        for coco_annotation in coco_annotation_handler.annotation_list:
            added_new, old_id, new_id = self.buffer.process_annotation(
                coco_annotation=coco_annotation, id_mapper=self.id_mapper, unique_key=ann_key
            )
            if verbose: print_id_update(added_new=added_new, old_id=old_id, new_id=new_id)

    def add_categories(self, ann_key: str, coco_category_handler: COCO_Category_Handler, verbose: bool=False):
        for coco_category in coco_category_handler.category_list:
            added_new, old_id, new_id = self.buffer.process_category(
                coco_category=coco_category, id_mapper=self.id_mapper, unique_key=ann_key
            )
            if verbose: print_id_update(added_new=added_new, old_id=old_id, new_id=new_id)

    def load_combined(self, update_img_paths: bool=True, verbose: bool=False, detailed_verbose: bool=False):
        for i, [img_dir, ann_path] in enumerate(zip(self.img_dir_list, self.ann_path_list)):
            if verbose: logger.info(f"{i}: Parsing {img_dir}, {ann_path}")
            parser = COCO_AnnotationFileParser(annotation_path=ann_path)
            parser.load(verbose=verbose)
            if verbose: logger.good("Annotations Loaded")
            self.add_licenses(ann_key=ann_path, coco_license_handler=parser.licenses, verbose=detailed_verbose)
            if verbose: logger.good(f"Licenses Updated")
            self.add_images(ann_key=ann_path, coco_image_handler=parser.images, img_dir=img_dir, update_img_paths=update_img_paths, verbose=detailed_verbose)
            if verbose: logger.good(f"Images Updated")
            self.add_categories(ann_key=ann_path, coco_category_handler=parser.categories, verbose=detailed_verbose)
            if verbose: logger.good(f"Categories Updated")
            self.add_annotations(ann_key=ann_path, coco_annotation_handler=parser.annotations, verbose=detailed_verbose)
            if verbose: logger.good(f"Annotations Updated")
        if verbose: logger.good(f"Finished Loaded Combined Annotations")

    def write_combined(self, output_path: str, verbose: bool=False):
        writer = COCO_Writer.from_buffer(buffer=self.buffer, output_path=output_path)
        json_dict = writer.build_json_dict(verbose=verbose)
        writer.write_json_dict(json_dict=json_dict, verbose=verbose)


from .objects import COCO_Info, COCO_License, COCO_Image, \
    COCO_Annotation, COCO_Category
from .handlers import COCO_License_Handler, COCO_Image_Handler, \
    COCO_Annotation_Handler, COCO_Category_Handler
from .dataset import COCO_Dataset